# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
"""
The ``Thumbs`` plugin provides a thumbnail index of all images viewed since
the program was started.

**Plugin Type: Global**

``Thumbs`` is a global plugin.  Only one instance can be opened.

**Usage**

By default, ``Thumbs`` appear in cronological viewing history,
with the newest images at the bottom and the oldest at the top.
The sorting can be made alphanumeric by a setting in the
"plugin_Thumbs.cfg" configuration file.

Clicking on a thumbnail navigates you directly to that image in the
associated channel.  Hovering the cursor over a thumbnail will show a
tool tip that contains a couple of useful pieces of metadata from the
image.

The "Auto Scroll" checkbox, if checked, will cause the ``Thumbs`` pan to
scroll to the active image.

This plugin is not usually configured to be closeable, but the user can
make it so by setting the "closeable" setting to True in the configuration
file--then Close and Help buttons will be added to the bottom of the UI.

**Excluding images from Thumbs**

.. note:: This also controls the behavior of ``Contents``.

Although the default behavior is for every image that is loaded into the
reference viewer to show up in ``Thumbs``, there may be cases where this
is undesirable (e.g., when there are many images being loaded at a
periodic rate by some automated process).  In such cases there are two
mechanisms for suppressing certain images from showing up in ``Thumbs``:

* Assigning the "genthumb" setting to False in a channel's settings
  (for example from the ``Preferences`` plugin, under the "General"
  settings) will exclude the channel itself and any of its images.
* Setting the "nothumb" keyword in the metadata of an image wrapper
  (not the FITS header, but by e.g., ``image.set(nothumb=True)``)
  will exclude that particular image from ``Thumbs``, even if the
  "genthumb" setting is True for that channel.

"""
import os
import math
import time
import threading
import bisect

from ginga import GingaPlugin
from ginga import RGBImage
from ginga.misc import Bunch
from ginga.util import iohelper
from ginga.gw import Widgets, Viewers
from ginga.util.paths import icondir
from ginga.pilw.ImageViewPil import CanvasView
from ginga.util.io import io_rgb

__all__ = ['Thumbs']


class Thumbs(GingaPlugin.GlobalPlugin):

    def __init__(self, fv):
        # superclass defines some variables for us, like logger
        super(Thumbs, self).__init__(fv)

        # For thumbnail pane
        self.thumblock = threading.RLock()
        self.thumb_dict = {}
        self.thumb_list = []
        self.thumb_num_cols = 1
        self._wd = 300
        self._ht = 400
        self._cmxoff = 0
        self._cmyoff = 0
        self._displayed_thumb_dict = {}
        tt_keywords = ['OBJECT', 'FRAMEID', 'UT', 'DATE-OBS']

        spec = self.fv.get_plugin_spec(str(self))

        prefs = self.fv.get_preferences()
        self.settings = prefs.create_category('plugin_Thumbs')
        self.settings.add_defaults(cache_thumbs=False,
                                   cache_location='local',
                                   auto_scroll=False,
                                   rebuild_wait=0.5,
                                   tt_keywords=tt_keywords,
                                   mouseover_name_key='NAME',
                                   thumb_length=180,
                                   thumb_hsep=15,
                                   thumb_vsep=15,
                                   sort_order=None,
                                   label_length=25,
                                   label_cutoff='right',
                                   highlight_tracks_keyboard_focus=True,
                                   label_font_color='white',
                                   label_font='sans condensed',
                                   label_font_size=10,
                                   label_bg_color='lightgreen',
                                   autoload_visible_thumbs=False,
                                   autoload_interval=0.25,
                                   update_interval=0.25,
                                   closeable=not spec.get('hidden', False),
                                   transfer_attrs=['transforms',
                                                   'cutlevels', 'rgbmap',
                                                   'icc', 'interpolation'])
        self.settings.load(onError='silent')
        # max length of thumb on the long side
        self.thumb_width = self.settings.get('thumb_length', 180)
        # distance in pixels between thumbs
        self.thumb_hsep = self.settings.get('thumb_hsep', 15)
        self.thumb_vsep = self.settings.get('thumb_vsep', 15)
        self.transfer_attrs = self.settings.get('transfer_attrs', [])

        # used to open thumbnails on disk
        self.rgb_opener = io_rgb.RGBFileHandler(self.logger)
        tmp_path = os.path.join(icondir, 'fits.png')
        self.placeholder_image = self.rgb_opener.load_file(tmp_path)

        # Build our thumb generator
        self.thumb_generator = self.get_thumb_generator()
        self.thumb_generator.name = 'thumb-generator'

        # a timer that controls how fast we attempt to update a thumbnail
        # after its associated full image has been modified
        self.timer_redo = self.fv.get_backend_timer()
        self.timer_redo.set_callback('expired', self.timer_redo_cb)
        self.lagtime = self.settings.get('rebuild_wait', 0.5)

        # a timer that controls how quickly we attempt to autoload missing
        # thumbnails
        self.timer_autoload = self.fv.get_backend_timer()
        self.timer_autoload.set_callback('expired', self.timer_autoload_cb)
        self.autoload_interval = self.settings.get('autoload_interval',
                                                   0.25)
        self.autoload_visible = self.settings.get('autoload_visible_thumbs',
                                                  False)
        self.autoload_serial = time.time()

        # timer that controls how quickly we attempt to rebuild thumbs after
        # a pan/scroll operation
        self.timer_update = self.fv.get_backend_timer()
        self.timer_update.set_callback('expired', self.timer_update_cb)
        self.update_interval = 0.25
        self._to_build = set([])
        self._latest_thumb = None
        self.save_thumbs = self.settings.get('cache_thumbs', False)

        # this will hold the thumbnails pane viewer
        self.c_view = None

        # TODO: these maybe should be configurable by channel
        # different instruments have different keywords of interest
        self.keywords = self.settings.get('tt_keywords', tt_keywords)
        self.keywords.insert(0, self.settings.get('mouseover_name_key', 'NAME'))

        self.re_hilite_set = set([])
        self.highlight_tracks_keyboard_focus = self.settings.get(
            'highlight_tracks_keyboard_focus', True)
        self._tkf_highlight = set([])

        fv.set_callback('add-image', self.add_image_cb)
        fv.set_callback('add-image-info', self.add_image_info_cb)
        fv.set_callback('remove-image-info', self.remove_image_info_cb)
        fv.set_callback('add-channel', self.add_channel_cb)
        fv.set_callback('delete-channel', self.delete_channel_cb)
        fv.add_callback('channel-change', self.focus_cb)

        self.gui_up = False

    def build_gui(self, container):
        vbox = Widgets.VBox()
        vbox.set_border_width(4)
        vbox.set_spacing(2)

        # construct an interactive viewer to view and scroll
        # the thumbs pane
        self.c_view = Viewers.CanvasView(logger=self.logger)
        c_v = self.c_view
        c_v.set_desired_size(self._wd, self._ht)
        c_v.enable_autozoom('off')
        c_v.enable_autocuts('off')
        c_v.set_pan(0, 0)
        c_v.scale_to(1.0, 1.0)
        # Y-axis flipped
        c_v.transform(False, True, False)
        c_v.cut_levels(0, 255)
        c_v.set_bg(0.4, 0.4, 0.4)
        # for debugging
        c_v.set_name('cmimage')
        c_v.add_callback('configure', self.thumbpane_resized_cb)
        c_v.add_callback('drag-drop', self.drag_drop_cb)
        c_v.get_settings().get_setting('pan').add_callback('set', self.thumbs_pan_cb)

        canvas = c_v.get_canvas()
        canvas.register_for_cursor_drawing(c_v)
        canvas.set_draw_mode('pick')
        canvas.ui_set_active(True, viewer=c_v)
        self.canvas = canvas

        bd = c_v.get_bindings()
        bd.enable_pan(True)
        bd.enable_zoom(False)
        bd.enable_cmap(False)
        bd.get_settings().set(scroll_pan_lock_x=True)

        # remap some bindings for pan mode into no mode needed
        bm = c_v.get_bindmap()
        for name in ['home', 'end', 'page_up', 'page_down',
                     'left', 'right', 'up', 'down']:
            bm.map_event(None, [], 'kp_%s' % name, 'pan_%s' % name)
        # scroll wheel
        bm.map_event(None, [], 'sc_scroll', 'pan')

        iw = Viewers.GingaScrolledViewerWidget(c_v)
        iw.resize(self._wd, self._ht)
        iw.scroll_bars(horizontal='auto', vertical='auto')

        vbox.add_widget(iw, stretch=1)

        captions = (('Auto scroll', 'checkbutton', 'Clear', 'button'),
                    )
        w, b = Widgets.build_info(captions)
        self.w.update(b)

        b.auto_scroll.set_tooltip(
            "Scroll the thumbs window when new images arrive")
        b.clear.set_tooltip("Remove all current thumbnails")
        b.clear.add_callback('activated', lambda w: self.clear())
        auto_scroll = self.settings.get('auto_scroll', False)
        b.auto_scroll.set_state(auto_scroll)
        vbox.add_widget(w, stretch=0)

        if self.settings.get('closeable', False):
            btns = Widgets.HBox()
            btns.set_border_width(4)
            btns.set_spacing(4)

            btn = Widgets.Button("Close")
            btn.add_callback('activated', lambda w: self.close())
            btns.add_widget(btn)
            btn = Widgets.Button("Help")
            btn.add_callback('activated', lambda w: self.help())
            btns.add_widget(btn, stretch=0)
            btns.add_widget(Widgets.Label(''), stretch=1)
            vbox.add_widget(btns, stretch=0)

        container.add_widget(vbox, stretch=1)

        self.gui_up = True

    def get_thumb_generator(self):
        tg = CanvasView(logger=self.logger)
        tg.configure_surface(self.thumb_width, self.thumb_width)
        tg.enable_autozoom('on')
        tg.set_autocut_params('histogram')
        tg.enable_autocuts('on')
        tg.enable_auto_orient(True)
        tg.defer_redraw = False
        tg.set_bg(0.7, 0.7, 0.7)
        return tg

    def drag_drop_cb(self, viewer, urls):
        """Punt drag-drops to the ginga shell.
        """
        channel = self.fv.get_current_channel()
        if channel is None:
            return
        self.fv.open_uris(urls, chname=channel.name, bulk_add=True)
        return True

    def start(self):
        self.add_visible_thumbs()

    def stop(self):
        self.gui_up = False

    def close(self):
        # clear current thumbs
        canvas = self.c_view.get_canvas()
        canvas.delete_all_objects()
        self._displayed_thumb_dict = dict()

        self.fv.stop_global_plugin(str(self))
        self.c_view = None
        return True

    def _get_thumb_key(self, chname, image):
        path = image.get('path', None)
        return self.get_thumb_key(chname, image.get('name'), path)

    def get_thumb_key(self, chname, imname, path):
        if path is not None:
            path = os.path.abspath(path)
        thumbkey = (chname, imname, path)
        return thumbkey

    def add_image_cb(self, fv, chname, image, info):
        """This callback happens when an image is loaded into one of the
        channels in the Ginga shell.
        """
        channel = self.fv.get_channel(chname)
        genthumb = channel.settings.get('genthumb', True)
        if not genthumb or image.get('nothumb', False):
            return False

        self.fv.gui_do(self._add_image, fv, chname, image, info)

    def _add_image(self, fv, chname, image, info):
        # invoked via add_image_cb()
        self.fv.assert_gui_thread()

        channel = self.fv.get_channel(chname)
        if info is None:
            try:
                imname = image.get('name', None)
                info = channel.get_image_info(imname)
            except KeyError:
                self.logger.warn("no information in channel about image '%s'" % (
                    imname))
                return False

        # Get any previously stored thumb information in the image info
        extras = info.setdefault('thumb_extras', Bunch.Bunch())

        # Get metadata for mouse-over tooltip
        metadata = self._get_tooltip_metadata(info, image)

        # Update the tooltip, in case of new or changed metadata
        text = self._mk_tooltip_text(metadata)
        extras.tooltip = text

        if 'rgbimg' not in extras:
            # since we have full size image in hand, generate a thumbnail
            # now, and cache it for when the thumb is added to the canvas
            thumb_image = self._regen_thumb_image(self.thumb_generator,
                                                  image, extras, channel.fitsimage)
            extras.rgbimg = thumb_image

        self._add_image_info(fv, channel, info)
        return True

    def add_image_info_cb(self, fv, channel, info):
        """This callback happens when an image is loaded into one of the
        channels in the Ginga shell OR information about an image (without)
        the actual image data is loaded into the channel (a lazy load).

        NOTE: in the case where image data is loaded into the channel, BOTH
        `add_image_cb` and `add_image_info_cb` will be called.
        """
        genthumb = channel.settings.get('genthumb', True)
        if not genthumb or info.get('nothumb', False):
            return False

        self.fv.gui_do(self._add_image_info, fv, channel, info)

    def _add_image_info(self, fv, channel, info):
        # invoked via add_image_info_cb()
        self.fv.assert_gui_thread()

        # Do we already have this thumb loaded?
        chname = channel.name
        thumbkey = self.get_thumb_key(chname, info.name, info.path)
        thumbpath = self.get_thumbpath(info.path)

        with self.thumblock:
            try:
                bnch = self.thumb_dict[thumbkey]

                if bnch.thumbpath == thumbpath:
                    self.logger.debug("we have this thumb--skipping regeneration")
                    return False

                # if these are not equal then the mtime must have
                # changed on the file, better reload and regenerate
                self.logger.debug("we have this thumb, but thumbpath is different--regenerating thumb")
                bnch.thumbpath = thumbpath
                #bnch.extras.setvals(placeholder=True)
                return

            except KeyError:
                self.logger.debug("we don't seem to have this thumb--generating thumb")

            self._insert_lazy_thumb(thumbkey, chname, info, thumbpath)

        return True

    def remove_image_info_cb(self, fv, channel, image_info):
        """This callback is called when an image is removed from a channel
        in the Ginga shell.
        """
        chname, imname, impath = channel.name, image_info.name, image_info.path
        try:
            thumbkey = self.get_thumb_key(chname, imname, impath)
            self.remove_thumb(thumbkey)
        except Exception as e:
            self.logger.error("Error removing thumb for %s: %s" % (
                imname, str(e)))

    def remove_thumb(self, thumbkey):
        """Remove the thumb indicated by `thumbkey`.
        """
        with self.thumblock:
            self.logger.debug("Removing thumb %s" % (str(thumbkey)))
            if thumbkey in self.thumb_dict:
                del self.thumb_dict[thumbkey]
            if thumbkey in self.thumb_list:
                self.thumb_list.remove(thumbkey)

            # Unhighlight
            chname = thumbkey[0]
            channel = self.fv.get_channel(chname)
            self._tkf_highlight.discard(thumbkey)
            channel.extdata.thumbs_old_highlight.discard(thumbkey)

        self.timer_update.set(self.update_interval)

    def thumbpane_resized_cb(self, thumb_viewer, width, height):
        """This callback is called when the Thumbs pane is resized.
        """
        self.fv.gui_do_oneshot('thumbs-resized', self._resized, width, height)

    def _resized(self, width, height):
        # invoked via thumbpane_resized_cb()
        self.fv.assert_gui_thread()
        self.logger.debug("thumbs resized, width=%d" % (width))

        with self.thumblock:
            self._cmxoff = -width // 2 + 10
            cols = max(1, width // (self.thumb_width + self.thumb_hsep))
            self.logger.debug("column count is now %d" % (cols))
            self.thumb_num_cols = cols

        self.timer_update.set(self.update_interval)
        return False

    def load_file(self, thumbkey, chname, info):
        """Called when a thumbnail is clicked (denoted by `thumbkey`)
        to load the thumbnail.
        """
        self.fv.assert_gui_thread()
        self.logger.debug("loading image: %s" % (str(thumbkey)))
        # TODO: deal with channel object directly?
        self.fv.switch_name(chname, info.name, path=info.path,
                            image_future=info.image_future)

    def clear(self):
        """Called when "Clear" button is clicked to clear the pane
        of thumbnails.
        """
        with self.thumblock:
            self.thumb_list = []
            self.thumb_dict = {}
            self._displayed_thumb_dict = {}
            self._tkf_highlight = set([])
            self.canvas.delete_all_objects(redraw=False)
            self.canvas.update_canvas(whence=0)

        self.timer_update.set(self.update_interval)

    def add_channel_cb(self, fv, channel):
        """Called when a channel is added from the main interface.
        Parameter is channel (a bunch).
        """
        fitsimage = channel.fitsimage
        fitssettings = fitsimage.get_settings()
        for name in ['cuts']:
            fitssettings.get_setting(name).add_callback(
                'set', self.cutset_cb, fitsimage)
        fitsimage.add_callback('transform', self.transform_cb)

        rgbmap = fitsimage.get_rgbmap()
        rgbmap.add_callback('changed', self.rgbmap_cb, fitsimage)

        # add old highlight set to channel external data
        channel.extdata.setdefault('thumbs_old_highlight', set([]))

    def focus_cb(self, fv, channel):
        """This callback is called when a channel viewer is focused.
        We use this to highlight the proper thumbs in the Thumbs pane.
        """
        # Reflect transforms, colormap, etc.
        fitsimage = channel.fitsimage
        image = channel.get_current_image()
        if image is not None:
            chname = channel.name
            thumbkey = self._get_thumb_key(chname, image)
            new_highlight = set([thumbkey])

            if self.have_thumbnail(fitsimage, image):
                # schedule an update of the thumbnail to pick up changes
                self.redo_delay(fitsimage)
        else:
            # no image has the focus
            new_highlight = set([])

        if self.highlight_tracks_keyboard_focus:
            self.update_highlights(self._tkf_highlight, new_highlight)
            self._tkf_highlight = new_highlight

    def transform_cb(self, fitsimage):
        """Called when a channel viewer has a transform event.
        Used to transform the corresponding thumbnail.
        """
        self.redo_delay(fitsimage)
        return True

    def cutset_cb(self, setting, value, fitsimage):
        """Called when a channel viewer has a cut levels event.
        Used to adjust cuts on the corresponding thumbnail.
        """
        self.redo_delay(fitsimage)
        return True

    def rgbmap_cb(self, rgbmap, fitsimage):
        """Called when a channel viewer has an RGB mapper event.
        Used to make color and contrast adjustments on the corresponding
        thumbnail.
        """
        # color mapping has changed in some way
        self.redo_delay(fitsimage)
        return True

    def redo_delay(self, fitsimage):
        # Delay regeneration of thumbnail until most changes have propagated
        self.timer_redo.data.setvals(fitsimage=fitsimage)
        self.timer_redo.set(self.lagtime)
        return True

    def timer_redo_cb(self, timer):
        """Called when the redo timer expires; used to rebuild the thumbnail
        corresponding to changes in the viewer.
        """
        self.fv.assert_gui_thread()
        self.redo_thumbnail(timer.data.fitsimage)

    def timer_autoload_cb(self, timer):
        """Called when the autoload timer expires; used to expand placeholder
        thumbnails.
        """
        self.fv.assert_gui_thread()
        self.fv.nongui_do(self._autoload_thumbs)

    def _autoload_thumbs(self):
        # invoked via timer_autoload_cb()
        self.logger.debug("autoload missing thumbs")
        self.fv.assert_nongui_thread()

        with self.thumblock:
            to_build = self._to_build.copy()

        serial = self.autoload_serial
        for thumbkey in to_build:
            if serial != self.autoload_serial or len(to_build) == 0:
                # cancel this autoload session if autoload set has changed
                return

            with self.thumblock:
                bnch = self.thumb_dict.get(thumbkey, None)
            if bnch is None:
                return

            # Get any previously stored thumb information in the image info
            extras = bnch.extras

            placeholder = extras.get('placeholder', True)
            ignore = extras.get('ignore', False)
            path = bnch.info.path
            if placeholder and path is not None and not ignore:
                self.force_load_for_thumb(thumbkey, path, bnch, extras)

    def force_load_for_thumb(self, thumbkey, path, bnch, extras):
        """Called by _autoload_thumbs() to load a file if the pane is
        currently showing a placeholder for a thumb.
        """
        self.logger.debug("autoload missing [%s]" % (path))
        info = bnch.info
        chname = thumbkey[0]
        channel = self.fv.get_channel(chname)
        try:
            thumb_image = self._get_thumb_image(channel, info, None, extras,
                                                no_placeholder=True)
            if thumb_image is not None:
                self.fv.gui_do(self.update_thumbnail, thumbkey, thumb_image,
                               None)
                return

            # <-- No easy thumb to load.  Forced to load the full image
            #     if we want a thumbnail
            if not self.autoload_visible:
                return

            image = self.fv.load_image(path, show_error=False)
            self.logger.debug("loaded [%s]" % (path))

            ## self.fv.gui_do(channel.add_image_update, image, info,
            ##                update_viewer=False)

            self.fv.gui_do(self.redo_thumbnail_image, channel, image, bnch,
                           save_thumb=self.save_thumbs)

        except Exception as e:
            self.logger.error("autoload missing [%s] failed:" % (path))
            # load errors will be reported in self.fv.load_image()
            # Just ignore autoload errors for now...
            extras.ignore = True

    def update_highlights(self, old_highlight_set, new_highlight_set):
        """Unhighlight the thumbnails represented by `old_highlight_set`
        and highlight the ones represented by new_highlight_set.

        Both are sets of thumbkeys.
        """
        with self.thumblock:
            common = old_highlight_set & new_highlight_set
            un_hilite_set = old_highlight_set - common
            re_hilite_set = new_highlight_set | common

            bg = self.settings.get('label_bg_color', 'lightgreen')
            fg = self.settings.get('label_font_color', 'white')

            # unhighlight thumb labels that should NOT be highlighted any more
            for thumbkey in un_hilite_set:
                if thumbkey in self.thumb_dict:
                    bnch = self.thumb_dict[thumbkey]
                    namelbl = bnch.get('namelbl', None)
                    if namelbl is not None:
                        namelbl.color = fg

            # highlight new labels that should be
            for thumbkey in re_hilite_set:
                if thumbkey in self.thumb_dict:
                    bnch = self.thumb_dict[thumbkey]
                    namelbl = bnch.get('namelbl', None)
                    if namelbl is not None:
                        namelbl.color = bg

            self.re_hilite_set = re_hilite_set

        if self.gui_up:
            self.c_view.redraw(whence=3)

    def redo(self, channel, image):
        """This method is called when an image is set in a channel.
        In this plugin it mostly serves to cause a different thumbnail
        label to be highlighted.
        """
        self.fv.assert_gui_thread()
        self.logger.debug("image set")
        chname = channel.name

        # get old highlighted thumbs for this channel -- will be
        # an empty set or one thumbkey
        old_highlight = channel.extdata.thumbs_old_highlight

        # calculate new highlight thumbkeys -- again, an empty set
        # or one thumbkey
        if image is not None:
            thumbkey = self._get_thumb_key(chname, image)
            new_highlight = set([thumbkey])
        else:
            # no image has the focus
            new_highlight = set([])

        nothumb = (image.get('nothumb', False) or
                   not channel.settings.get('genthumb', True))

        with self.thumblock:
            #if not self.have_thumbnail(channel.fitsimage, image):
            if thumbkey not in self.thumb_dict and not nothumb:
                # No memory of this thumbnail, so regenerate it
                if not self._add_image(self.fv, chname, image, None):
                    return

        # this would have auto scroll feature pan to most recent image
        # loaded in channel
        #self.auto_scroll(thumbkey)

        self.logger.debug("highlighting")
        # Only highlights active image in the current channel
        if self.highlight_tracks_keyboard_focus:
            self.update_highlights(self._tkf_highlight, new_highlight)
            self._tkf_highlight = new_highlight

        # Highlight all active images in all channels
        else:
            self.update_highlights(old_highlight, new_highlight)
            channel.extdata.thumbs_old_highlight = new_highlight

        self.redo_delay(channel.fitsimage)

    def have_thumbnail(self, fitsimage, image):
        """Returns True if we already have a thumbnail version of this image
        cached, False otherwise.
        """
        chname = self.fv.get_channel_name(fitsimage)

        # Look up our version of the thumb
        idx = image.get('idx', None)
        path = image.get('path', None)
        if path is not None:
            path = os.path.abspath(path)
            name = iohelper.name_image_from_path(path, idx=idx)
        else:
            name = 'NoName'

        # get image name
        name = image.get('name', name)

        thumbkey = self.get_thumb_key(chname, name, path)
        with self.thumblock:
            return thumbkey in self.thumb_dict

    def _save_thumb(self, thumb_image, bnch):
        extras = bnch.extras
        if extras.get('ignore', False) or extras.get('placeholder', False):
            # don't save placeholders, or thumbs we are instructed to ignore
            return

        thumbpath = self.get_thumbpath(bnch.info.path)
        if thumbpath is not None:
            if os.path.exists(thumbpath):
                os.remove(thumbpath)
            thumb_image.save_as_file(thumbpath)

    def redo_thumbnail(self, viewer, save_thumb=None):
        """Regenerate the thumbnail for the image in the channel viewer
        (`viewer`).  If `save_thumb` is `True`, then save a copy of the
        thumbnail if the user has allowed it.
        """
        # Get the thumbnail image
        image = viewer.get_image()
        if image is None:
            return

        imname = image.get('name', None)
        if imname is None:
            return

        chname = self.fv.get_channel_name(viewer)
        channel = self.fv.get_channel(chname)
        try:
            info = channel[imname]
        except KeyError:
            # don't generate a thumbnail without info
            return

        thumbkey = self.get_thumb_key(chname, imname, info.path)
        with self.thumblock:
            if thumbkey not in self.thumb_dict:
                # No memory of this thumbnail
                return

            bnch = self.thumb_dict[thumbkey]
            self.redo_thumbnail_image(channel, image, bnch,
                                      save_thumb=save_thumb)

    def redo_thumbnail_image(self, channel, image, bnch, save_thumb=None):
        """Regenerate the thumbnail for image `image`, in the channel
        `channel` and whose entry in the thumb_dict is `bnch`.
        If `save_thumb` is `True`, then save a copy of the thumbnail if
        the user has allowed it.
        """
        # image is flagged not to make a thumbnail?
        nothumb = (image.get('nothumb', False) or
                   not channel.settings.get('genthumb', True))
        if nothumb:
            return

        self.logger.debug("redoing thumbnail ...")
        if save_thumb is None:
            save_thumb = self.save_thumbs

        # Get any previously stored thumb information in the image info
        extras = bnch.extras

        # Get metadata for mouse-over tooltip
        info = bnch.info
        metadata = self._get_tooltip_metadata(info, image)

        chname = channel.name
        thumbkey = self.get_thumb_key(chname, info.name, info.path)
        with self.thumblock:
            if thumbkey not in self.thumb_dict:
                # This shouldn't happen, but if it does, ignore the thumbkey
                self.logger.warning("No memory of %s..." % (str(thumbkey)))
                return

            bnch = self.thumb_dict[thumbkey]

            # Generate new thumbnail
            self.logger.debug("generating new thumbnail")
            thumb_image = self._regen_thumb_image(self.thumb_generator,
                                                  image, extras, channel.fitsimage)

            # Save a thumbnail for future browsing
            if save_thumb and info.path is not None:
                try:
                    self._save_thumb(thumb_image, bnch)
                except Exception as e:
                    # if we can't persist the thumbnail, don't let that
                    # stop everything
                    self.logger.error("Couldn't persist thumbnail: {}".format(e))

            self.update_thumbnail(thumbkey, thumb_image, metadata)

        self.fv.update_pending()

    def delete_channel_cb(self, fv, channel):
        """Called when a channel is deleted from the main interface.
        Parameter is channel (a bunch).
        """
        chname_del = channel.name
        # TODO: delete thumbs for this channel!
        self.logger.debug("deleting thumbs for channel '%s'" % (chname_del))
        with self.thumblock:
            new_thumb_list = []
            un_hilite_set = set([])
            for thumbkey in self.thumb_list:
                chname = thumbkey[0]
                if chname != chname_del:
                    new_thumb_list.append(thumbkey)
                else:
                    if thumbkey in self.thumb_dict:
                        del self.thumb_dict[thumbkey]
                    un_hilite_set.add(thumbkey)

            self.thumb_list = new_thumb_list
            self._tkf_highlight -= un_hilite_set  # Unhighlight

        self.timer_update.set(self.update_interval)

    def _get_tooltip_metadata(self, info, image, keywords=None):
        """Construct a metadata dict containing values for selected
        `keywords` (defaults to `self.keywords`) for the image `image`
        (can be `None`) whose info dict is `info`.
        """
        # Get metadata for mouse-over tooltip
        header = {}
        if image is not None:
            header = image.get_header()

        if keywords is None:
            keywords = self.keywords
        metadata = {kwd: header.get(kwd, 'N/A')
                    for kwd in keywords}

        # assign a name in the metadata if we don't have one yet
        name_key = self.settings.get('mouseover_name_key', 'NAME')
        if metadata.setdefault(name_key, info.name) == 'N/A':
            metadata[name_key] = info.name

        return metadata

    def make_tt(self, thumbs_viewer, canvas, text, pt, obj, fontsize=10):
        """Create a tooltip pop-up for a thumbnail to be displayed in
        `thumbs_viewer` on its `canvas` based on `text` where the cursor
        is at point `pt` and on the thumbnail object `obj`.  This object
        is created but not added to the canvas.

        NOTE: currently `pt` is not used, but kept for possible future use.
        """
        # Determine pop-up position on canvas.  Try to align a little below
        # the thumbnail image and offset a bit
        x1, y1, x2, y2 = obj.get_llur()

        x = x1 + 10
        y = y1 + 10
        mxwd = 0
        lines = text.split('\n')

        # create the canvas object representing the tooltip
        point = canvas.dc.Point(x, y, radius=0, color='black', alpha=0.0)
        rect = canvas.dc.Rectangle(x, y, x, y, color='black', fill=True,
                                   fillcolor='lightyellow')
        crdmap = thumbs_viewer.get_coordmap('offset')
        crdmap.refobj = point

        l = [point, rect]
        a, b = 2, 0
        for line in lines:
            text = canvas.dc.Text(a, b, text=line, color='black',
                                  font='sans condensed', fontsize=fontsize)
            text.crdmap = crdmap
            l.append(text)
            txt_wd, txt_ht = thumbs_viewer.renderer.get_dimensions(text)
            b += txt_ht + 2
            text.y = b
            mxwd = max(mxwd, txt_wd)
        rect.x2 = rect.x1 + mxwd + 2
        rect.y2 = rect.y1 + b + 4

        obj = canvas.dc.CompoundObject(*l)

        # sanity check and adjustment so that popup will be minimally obscured
        # by a window edge
        x3, y3, x4, y4 = thumbs_viewer.get_data_rect()
        if rect.x2 > x4:
            off = rect.x2 - x4
            rect.x1 -= off
            rect.x2 -= off
            point.x -= off
        if rect.y1 < y3:
            off = y3 - rect.y1
            rect.y1 += off
            rect.y2 += off
            point.y += off
        if rect.y2 > y4:
            off = rect.y2 - y4
            rect.y1 -= off
            rect.y2 -= off
            point.y -= off

        return obj

    def show_tt(self, obj, canvas, event, pt,
                thumbkey, chname, info, tf):
        """Removing any previous tooltip, and if `tf` is `True`, pop up a
        new tooltip on the thumbnail viewer window for the thumb denoted
        by `thumbkey`.  `obj` is the thumbnail compound object, `canvas`
        if the thumbnail viewer canvas, `event` is the cursor event that
        caused the popup. The cursor is at point `pt`. The channel for this
        image has name `chname` and has image info `info`.
        """
        text = info.thumb_extras.tooltip

        tag = '_$tooltip'
        try:
            canvas.delete_object_by_tag(tag)
        except KeyError:
            pass
        if tf:
            tt = self.make_tt(self.c_view, canvas, text, pt, obj)
            canvas.add(tt, tag=tag)

    def _regen_thumb_image(self, tg, image, extras, viewer):
        self.logger.debug("generating new thumbnail")

        if not tg.viewable(image):
            # this is not something we know how to open
            # TODO: other viewers might be able to open it, need to check
            # with them
            image = self.placeholder_image
            extras.setvals(rgbimg=image, placeholder=False,
                           ignore=True, time_update=time.time())
            return image

        tg.set_image(image)
        if viewer is not None:
            # if a viewer was passed, and there is an image loaded there,
            # then copy the viewer attributes to the thumbnail generator
            v_img = viewer.get_image()
            if v_img is not None:
                # this should force thumb generator to redraw as needed
                viewer.copy_attributes(tg, self.transfer_attrs)

        order = tg.rgb_order
        rgb_img = tg.get_image_as_array(order=order)
        thumb_image = RGBImage.RGBImage(rgb_img, order=order)
        extras.setvals(rgbimg=thumb_image, placeholder=False,
                       time_update=time.time())
        return thumb_image

    def _get_thumb_image(self, channel, info, image, extras,
                         no_placeholder=False):
        """Get a thumb image for the image `image` (can be `None`) that
        is associated with channel `channel` and image information `info`.
        """

        # Choice [A]: is there a thumb image attached to the image info?
        if 'rgbimg' in extras:
            # yes
            return extras.rgbimg

        thumbpath = self.get_thumbpath(info.path)

        # Choice [B]: is the full image available to make a thumbnail?
        if image is None:
            try:
                image = channel.get_loaded_image(info.name)

            except KeyError:
                pass

        if image is not None:
            try:
                thumb_image = self._regen_thumb_image(self.thumb_generator,
                                                      image, extras,
                                                      channel.fitsimage)
                extras.setvals(rgbimg=thumb_image, placeholder=False,
                               time_update=None)
                return thumb_image

            except Exception as e:
                self.logger.warning("Error generating thumbnail: %s" % (str(e)))

        # Choice [C]: is there a cached thumbnail image on disk we can use?
        if (thumbpath is not None) and os.path.exists(thumbpath):
            try:
                # try to load the thumbnail image
                thumb_image = self.rgb_opener.load_file(thumbpath)
                wd, ht = thumb_image.get_size()[:2]
                if max(wd, ht) > self.thumb_width:
                    # <-- thumb size does not match our expected size
                    thumb_image = self._regen_thumb_image(self.thumb_generator,
                                                          thumb_image, extras,
                                                          None)
                thumb_image.set(name=info.name)
                extras.setvals(rgbimg=thumb_image, placeholder=False,
                               time_update=None)
                return thumb_image

            except Exception as e:
                self.logger.warning("Error loading thumbnail: %s" % (str(e)))

        if no_placeholder:
            return None

        # Choice [D]: use a placeholder image
        data_np = self.placeholder_image.get_data()
        thumb_image = RGBImage.RGBImage(data_np=data_np)
        thumb_image.set(name=info.name, path=None)
        extras.setvals(placeholder=True, time_update=None)

        return thumb_image

    def get_thumbpath(self, path, makedir=True):
        """Return the path for the thumbnail location on disk based on the
        path of the original.  Can return `None` if there is no suitable path.
        The preference for where to store thumbnails is set in the settings
        for this plugin.
        """
        if path is None:
            return None

        path = os.path.abspath(path)
        dirpath, filename = os.path.split(path)
        # Get thumb directory
        cache_location = self.settings.get('cache_location', 'local')
        if cache_location == 'ginga':
            # thumbs in .ginga cache
            prefs = self.fv.get_preferences()
            thumbdir = os.path.join(prefs.get_baseFolder(), 'thumbs')
            thumbdir = os.path.join(thumbdir, iohelper.gethex(dirpath))
        else:
            # thumbs in .thumbs subdirectory of image folder
            thumbdir = os.path.join(dirpath, '.thumbs')

        try:
            thumbpath = iohelper.get_thumbpath(thumbdir, path,
                                               makedir=makedir)

        except Exception as e:
            self.logger.debug("Error getting thumb path for '%s': %s" % (
                path, str(e)))
            thumbpath = None

        return thumbpath

    def _calc_thumb_pos(self, row, col):
        """Calculate the thumb data coordinates on the thumbs canvas for
        a thumb at row `row` and column `col`.  Returns a 4-tuple of x/y
        positions for the text item and the image item.
        """
        self.logger.debug("row, col = %d, %d" % (row, col))
        # TODO: should really be text ht
        text_ht = self.thumb_vsep

        # calc in data coords
        thumb_height = self.thumb_width
        twd_hplus = self.thumb_width + self.thumb_hsep
        twd_vplus = thumb_height + text_ht + self.thumb_vsep
        xt = self._cmxoff + (col * twd_hplus)
        yt = self._cmyoff + (row * twd_vplus)

        # position of image
        xi = xt
        yi = yt + 6
        self.logger.debug("xt, yt = %d, %d" % (xt, yt))

        return (xt, yt, xi, yi)

    def get_visible_rows(self):
        """Return the list of thumbkeys for the thumbs that should be
        visible in the Thumbs pane.
        """
        x1, y1, x2, y2 = self.c_view.get_data_rect()
        self.logger.debug("datarect=(%f, %f, %f, %f)", x1, y1, x2, y2)

        # TODO: should really be text ht
        text_ht = self.thumb_vsep

        # calc in data coords
        thumb_height = self.thumb_width
        twd_vplus = thumb_height + text_ht + self.thumb_vsep
        # remember, Y-axis is flipped
        row1 = max(0, int(math.floor(y1 / twd_vplus) - 2))
        row2 = max(row1 + 1, int(math.ceil(y2 / twd_vplus) + 2))
        self.logger.debug("row1, row2 = %d, %d" % (row1, row2))
        return row1, row2

    def get_visible_thumbs(self):
        """Return the list of thumbkeys for the thumbs that should be
        visible in the Thumbs pane.
        """
        row1, row2 = self.get_visible_rows()
        i = max(0, row1 * self.thumb_num_cols)
        j = min(len(self.thumb_list) - 1, row2 * self.thumb_num_cols)
        self.logger.debug("i, j = %d, %d" % (i, j))
        thumbs = [self.thumb_list[n] for n in range(i, j + 1)]
        return thumbs

    def add_visible_thumbs(self):
        """Add the thumbs to the canvas that should be visible in the
        Thumbs pane and remove the others.
        """
        if not self.gui_up:
            return
        self.fv.assert_gui_thread()

        canvas = self.c_view.get_canvas()

        with self.thumblock:
            thumb_keys = set(self.get_visible_thumbs())
            old_thumb_keys = set(self._displayed_thumb_dict.keys())
            if old_thumb_keys == thumb_keys:
                # no need to do anything
                self.c_view.redraw(whence=0)
                return

            to_delete = old_thumb_keys - thumb_keys
            to_add = thumb_keys - old_thumb_keys

            # make a copy of these for potential building thumbs
            self._to_build = set(thumb_keys)

            # delete thumbs from canvas that are no longer visible
            canvas.delete_all_objects(redraw=False)
            ## for thumbkey in to_delete:
            ##     bnch = self._displayed_thumb_dict[thumbkey]
            ##     if 'widget' in bnch and canvas.has_object(bnch.widget):
            ##         canvas.delete_object(bnch.widget, redraw=False)

            # update displayed thumbs dict
            self._displayed_thumb_dict = {thumbkey: self.thumb_dict[thumbkey]
                                          for thumbkey in thumb_keys}

            # add newly-visible thumbs to canvas
            row1, row2 = self.get_visible_rows()
            row, col = row1, 0

            while row <= row2:
                i = row * self.thumb_num_cols + col
                if 0 <= i < len(self.thumb_list):
                    thumbkey = self.thumb_list[i]
                    bnch = self.thumb_dict[thumbkey]
                    bnch.setvals(row=row, col=col)

                    if 'widget' not in bnch:
                        # lazily make a canvas object for "empty" thumb entries
                        self._make_thumb(bnch, thumbkey)
                    else:
                        # update object positions for existing entries
                        xt, yt, xi, yi = self._calc_thumb_pos(row, col)
                        bnch.namelbl.x, bnch.namelbl.y = xt, yt
                        bnch.image.x, bnch.image.y = xi, yi

                    if not canvas.has_object(bnch.widget):
                        canvas.add(bnch.widget, redraw=False)

                col = (col + 1) % self.thumb_num_cols
                if col == 0:
                    row += 1

        self.c_view.redraw(whence=0)

        self.fv.update_pending()

        # load and create thumbnails for any placeholder icons
        self.autoload_serial = time.time()
        #self.fv.gui_do_oneshot('thumbs-autoload', self._autoload_thumbs)
        self.timer_autoload.set(self.autoload_interval)

    def update_thumbs(self):
        """This is called whenever the thumb list has changed.
        """
        self.fv.assert_gui_thread()

        with self.thumblock:
            n = len(self.thumb_list)
            row = n // self.thumb_num_cols
            col = n % self.thumb_num_cols

            thumbkey = None
            if self._latest_thumb in self.thumb_dict:
                thumbkey = self._latest_thumb
            else:
                self._latest_thumb = None

        # update the visible limits (e.g. scroll bars)
        xm, ym, x_, y_ = self._calc_thumb_pos(0, 0)
        xt, yt, xi, yi = self._calc_thumb_pos(row, col)
        xi += self.thumb_width * 2
        self.c_view.set_limits([(xm, ym), (xi, yi)], coord='data')

        if thumbkey is not None:
            self.auto_scroll(thumbkey)

        # update the thumbs pane in the case that thumbs visibility
        # has changed
        self.add_visible_thumbs()

    def thumbs_pan_cb(self, thumbs_viewer, pan_vec):
        """This callback is called when the Thumbs pane is panned/scrolled.
        """
        self.fv.gui_do_oneshot('thumbs-pan', self.add_visible_thumbs)

    def _insert_lazy_thumb(self, thumbkey, chname, info, thumbpath):
        """This function gets called to create an initial entry for a
        thumb.
        """
        thumbname = info.name
        self.logger.debug("inserting an empty thumb %s" % (thumbname))

        # Make an initial entry for the thumbs in the tracking dict.
        # Nothing is actually plotted, because the thumb may be in a region
        # that is not visible.
        n = len(self.thumb_list)
        row = n // self.thumb_num_cols
        col = n % self.thumb_num_cols

        extras = info.setdefault('thumb_extras', Bunch.Bunch())
        bnch = Bunch.Bunch(info=info, extras=extras,
                           thumbname=thumbname, chname=chname,
                           thumbpath=thumbpath, row=row, col=col)

        self.thumb_dict[thumbkey] = bnch
        if thumbkey not in self.thumb_list:
            sort_order = self.settings.get('sort_order', None)
            if sort_order is not None:
                bisect.insort(self.thumb_list, thumbkey)
            else:
                self.thumb_list.append(thumbkey)

        self._latest_thumb = thumbkey

        self.timer_update.cond_set(self.update_interval)
        #self.timer_update.set(self.update_interval)

    def _make_thumb(self, bnch, thumbkey):
        """Called from inside add_visible_thumbs() to generate a placeholder
        thumb for thumbs that are not yet made manifest.
        """
        channel = self.fv.get_channel(bnch.chname)
        info = bnch.info
        image = None

        # Get any previously stored thumb information in the image info
        extras = bnch.extras

        thumbpath = self.get_thumbpath(info.path)
        thumb_image = extras.get('rgbimg', None)
        if thumb_image is None:
            thumb_image = self._get_thumb_image(channel, info, image, extras)

        # Get metadata for mouse-over tooltip
        metadata = self._get_tooltip_metadata(info, None)

        thumbname = info.name
        self.logger.debug("inserting thumb %s" % (thumbname))
        chname = bnch.chname

        # If there is no previously made tooltip, then generate one
        if 'tooltip' in extras:
            text = extras.tooltip
        else:
            text = self._mk_tooltip_text(metadata)
            extras.tooltip = text

        canvas = self.canvas
        fg = self.settings.get('label_font_color', 'white')
        bg = self.settings.get('label_bg_color', 'lightgreen')
        fontname = self.settings.get('label_font', 'sans condensed')
        fontsize = self.settings.get('label_font_size', 10)

        # Shorten thumbnail label, if requested
        label_length = self.settings.get('label_length', None)
        label_cutoff = self.settings.get('label_cutoff', None)

        if label_length is not None:
            if label_cutoff is not None:
                thumbname = iohelper.shorten_name(thumbname, label_length,
                                                  side=label_cutoff)
            else:
                thumbname = thumbname[:label_length]

        row, col = bnch.row, bnch.col

        xt, yt, xi, yi = self._calc_thumb_pos(row, col)
        l2 = []
        color = bg if thumbkey in self.re_hilite_set else fg
        namelbl = canvas.dc.Text(xt, yt, text=thumbname, color=color,
                                 font=fontname, fontsize=fontsize,
                                 coord='data')
        l2.append(namelbl)

        image = canvas.dc.Image(xi, yi, thumb_image, alpha=1.0,
                                linewidth=1, color='black', coord='data')
        l2.append(image)

        obj = canvas.dc.CompoundObject(*l2, coord='data')
        obj.pickable = True
        obj.opaque = True

        bnch.setvals(widget=obj, image=image, namelbl=namelbl,
                     thumbpath=thumbpath)

        # set the load callback
        obj.add_callback('pick-down',
                         lambda *args: self.load_file(thumbkey, chname,
                                                      info))
        # set callbacks for tool tips
        obj.add_callback('pick-enter', self.show_tt,
                         thumbkey, chname, info, True)
        obj.add_callback('pick-leave', self.show_tt,
                         thumbkey, chname, info, False)

    def auto_scroll(self, thumbkey):
        """Scroll the Thumbs viewer to the thumb denoted by `thumbkey`.
        """
        if not self.gui_up:
            return
        # force scroll to bottom of thumbs, if checkbox is set
        scrollp = self.w.auto_scroll.get_state()
        if not scrollp:
            return

        with self.thumblock:
            i = self.thumb_list.index(thumbkey)

        row = i // self.thumb_num_cols
        col = i % self.thumb_num_cols

        # override X parameter because we only want to scroll vertically
        pan_x, pan_y = self.c_view.get_pan()
        xt, yt, xi, yi = self._calc_thumb_pos(row, col)
        self.c_view.panset_xy(pan_x, yi)

    def clear_widget(self):
        """Clears the thumbnail viewer of all thumbnails, but does
        not remove them from the thumb_dict or thumb_list.
        """
        if not self.gui_up:
            return
        canvas = self.c_view.get_canvas()
        canvas.delete_all_objects()
        self.c_view.redraw(whence=0)

    def timer_update_cb(self, timer):
        if not self.gui_up:
            return
        self.fv.gui_do_oneshot('thumbs-update', self.update_thumbs)

    def _mk_tooltip_text(self, metadata):
        """Make a tooltip text from values related to interesting keywords
        in `metadata`.
        """
        result = []
        for kwd in self.keywords:
            try:
                text = kwd + ': ' + str(metadata[kwd])

            except Exception as e:
                self.logger.debug("Couldn't get keyword '%s' value: %s" % (
                    kwd, str(e)))
                text = "%s: N/A" % (kwd)
            result.append(text)

        return '\n'.join(result)

    def update_thumbnail(self, thumbkey, thumb_image, metadata):
        """Update the thumbnail denoted by `thumbkey` with a new thumbnail
        image (`thumb_image`) and new knowledge of dict `metadata`.
        """
        with self.thumblock:
            try:
                bnch = self.thumb_dict[thumbkey]
            except KeyError:
                self.logger.debug("No thumb found for %s; not updating "
                                  "thumbs" % (str(thumbkey)))
                return

            info = bnch.info
            # Get any previously stored thumb information in the image info
            extras = bnch.extras

            # Update the tooltip, in case of new or changed metadata
            if metadata is not None:
                text = self._mk_tooltip_text(metadata)
                extras.tooltip = text

            self.logger.debug("updating thumbnail '%s'" % (info.name))
            extras.rgbimg = thumb_image
            cvs_img = bnch.get('image', None)
            if cvs_img is not None:
                # TODO: figure out why set_image() causes corruption of the
                # redraw here.  Instead we force a manual redraw.
                #image.set_image(thumb_image)
                cvs_img.image = thumb_image
                # TODO: need to set width, height?

            if self.gui_up and thumbkey in self._displayed_thumb_dict:
                # redraw the thumbs viewer if the update was to a displayed
                # thumb
                self.c_view.redraw(whence=0)
            self.logger.debug("update finished.")

    def __str__(self):
        return 'thumbs'


# Append module docstring with config doc for auto insert by Sphinx.
from ginga.util.toolbox import generate_cfg_example  # noqa
if __doc__ is not None:
    __doc__ += generate_cfg_example('plugin_Thumbs', package='ginga')

# END



#! /usr/bin/env python
#
# example2.py -- Simple, configurable FITS viewer.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#

import sys

from ginga.gtk4w import GtkHelp
from ginga.gtk4w.ImageViewGtk import CanvasView, ScrolledView
from ginga.canvas.CanvasObject import get_canvas_types
from ginga import colors
from ginga.misc import log
from ginga.util.loader import load_data
from ginga.locale.localize import _tr

from gi.repository import Gtk
#from gi.repository import GLib

STD_FORMAT = '%(asctime)s | %(levelname)1.1s | %(filename)s:%(lineno)d (%(funcName)s) | %(message)s'


class FitsViewer(object):

    def __init__(self, logger, render='widget'):
        self.logger = logger
        self.render = render
        self.app = Gtk.Application()
        self.app.connect('activate', self.on_activate_cb)

        self.gui_up = False

    def on_activate_cb(self, app):
        root = Gtk.ApplicationWindow(application=app)
        root.set_title("Gtk4 CanvasView Example")
        #root.set_border_width(2)
        #root.connect("delete_event", lambda w, e: quit(w))
        self.root = root
        self.select = GtkHelp.FileSelection(root)

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        vbox.set_spacing(2)

        self.drawcolors = colors.get_colors()
        self.dc = get_canvas_types()

        fi = CanvasView(self.logger, render=self.render)
        fi.enable_autocuts('on')
        fi.set_autocut_params('zscale')
        fi.enable_autozoom('on')
        fi.set_zoom_algorithm('rate')
        fi.set_zoomrate(1.4)
        fi.show_pan_mark(True)
        fi.set_callback('drag-drop', self.drop_file_cb)
        fi.set_callback('cursor-changed', self.cursor_cb)
        fi.set_bg(0.2, 0.2, 0.2)
        fi.ui_set_active(True)
        fi.set_enter_focus(True)
        # add little mode indicator that shows keyboard modal states
        fi.show_mode_indicator(True, corner='ur')
        # add a color bar
        #fi.show_color_bar(True)
        #fi.show_focus_indicator(True)
        self.viewer = fi

        bd = fi.get_bindings()
        bd.enable_all(True)

        # canvas that we will draw on
        canvas = self.dc.DrawingCanvas()
        canvas.enable_draw(True)
        canvas.set_drawtype('rectangle', color='lightblue')
        canvas.register_for_cursor_drawing(fi)
        canvas.set_draw_mode('draw')
        canvas.set_surface(fi)
        canvas.ui_set_active(True)
        self.canvas = canvas

        # add our new canvas to viewers default canvas
        fi.get_canvas().add(canvas)

        self.drawtypes = canvas.get_drawtypes()
        self.drawtypes.sort()

        try:
            w = self.viewer.get_widget()
            w.set_size_request(512, 512)

            # add scrollbar interface around this viewer
            si = ScrolledView(fi)
            si.scroll_bars(horizontal='auto', vertical='auto')

            vbox.append(si)

            self.readout = Gtk.Label(label="")
            #self.readout.hexpand(True)
            vbox.append(self.readout)

            hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
            hbox.set_spacing(5)
            hbox.set_margin_start(4)
            hbox.set_margin_end(4)
            hbox.set_margin_top(4)
            hbox.set_margin_bottom(4)

            wdrawtype = GtkHelp.combo_box_new_text()
            index = 0
            for name in self.drawtypes:
                wdrawtype.insert_text(index, name)
                index += 1
            index = self.drawtypes.index('rectangle')
            wdrawtype.set_active(index)
            wdrawtype.connect('changed', self.set_drawparams)
            self.wdrawtype = wdrawtype

            wdrawcolor = GtkHelp.combo_box_new_text()
            index = 0
            for name in self.drawcolors:
                wdrawcolor.insert_text(index, name)
                index += 1

            index = self.drawcolors.index('lightblue')
            wdrawcolor.set_active(index)
            wdrawcolor.connect('changed', self.set_drawparams)
            self.wdrawcolor = wdrawcolor

            wfill = GtkHelp.CheckButton(label=_tr("Fill"))
            wfill.sconnect('toggled', self.set_drawparams)
            self.wfill = wfill

            walpha = GtkHelp.SpinButton()
            adj = walpha.get_adjustment()
            adj.configure(0.0, 0.0, 1.0, 0.1, 0.1, 0)
            walpha.set_value(1.0)
            walpha.set_digits(1)
            walpha.sconnect('value-changed', self.set_drawparams)
            self.walpha = walpha

            wclear = Gtk.Button(label=_tr("Clear Canvas"))
            wclear.connect('clicked', self.clear_canvas)

            wopen = Gtk.Button(label=_tr("Open File"))
            wopen.connect('clicked', self.open_file)
            wquit = Gtk.Button(label=_tr("Quit"))
            wquit.connect('clicked', quit)

            for w in (wquit, wclear, walpha, Gtk.Label(label=_tr("Alpha") + ":"),
                      wfill, wdrawcolor, wdrawtype, wopen):
                hbox.prepend(w)

            vbox.append(hbox)

            root.set_child(vbox)
            root.present()

            print("finishing gui")
            self.gui_up = True
        except Exception as e:
            self.logger.error(f"error bringing up GUI: {e}", exc_info=True)
            #sys.exit(1)

        return True

    def get_widget(self):
        return self.root

    def set_drawparams(self, w):
        if not self.gui_up:
            return
        index = self.wdrawtype.get_active()
        kind = self.drawtypes[index]
        index = self.wdrawcolor.get_active()
        fill = self.wfill.get_active()
        alpha = self.walpha.get_value()

        params = {'color': self.drawcolors[index],
                  'alpha': alpha,
                  #'cap': 'ball',
                  }
        if kind in ('circle', 'rectangle', 'polygon', 'triangle',
                    'righttriangle', 'ellipse', 'square', 'box'):
            params['fill'] = fill
            params['fillalpha'] = alpha

        self.canvas.set_drawtype(kind, **params)

    def clear_canvas(self, w):
        self.canvas.delete_all_objects()

    def load_file(self, filepath):
        image = load_data(filepath, logger=self.logger)
        self.viewer.set_image(image)
        self.root.set_title(filepath)

    def open_file(self, w):
        self.select.popup(_tr("Open FITS file"), self.load_file)

    def drop_file_cb(self, fitsimage, paths):
        fileName = paths[0]
        self.load_file(fileName)

    def cursor_cb(self, viewer, button, data_x, data_y):
        """This gets called when the data position relative to the cursor
        changes.
        """
        # Get the value under the data coordinates
        try:
            # We report the value across the pixel, even though the coords
            # change halfway across the pixel
            value = viewer.get_data(int(data_x + viewer.data_off),
                                    int(data_y + viewer.data_off))

        except Exception:
            value = None

        fits_x, fits_y = data_x + 1, data_y + 1

        # Calculate WCS RA
        try:
            # NOTE: image function operates on DATA space coords
            image = viewer.get_image()
            if image is None:
                # No image loaded
                return
            ra_txt, dec_txt = image.pixtoradec(fits_x, fits_y,
                                               format='str', coords='fits')
        except Exception as e:
            self.logger.warning(_tr("Bad coordinate conversion") + ": %s" % (
                str(e)))
            ra_txt = _tr('BAD WCS')
            dec_txt = _tr('BAD WCS')

        text = "RA: %s  DEC: %s  X: %.2f  Y: %.2f  Value: %s" % (
            ra_txt, dec_txt, fits_x, fits_y, value)
        self.readout.set_text(text)

    def quit(self, w):
        self.app.quit()
        return True

    # def mainloop(self):
    #     context = GLib.main_context_default()
    #     while True:
    #         while context.pending():
    #             try:
    #                 context.iteration()
    #             except Exception as e:
    #                 self.logger.error("Exception in main_iteration() loop: %s" %
    #                                   (str(e)))

    def mainloop(self):
        self.app.run(None)


def main(options, args):

    logger = log.get_logger("example2", options=options)

    print('example RENDER  is', options.render)
    fv = FitsViewer(logger, render=options.render)
    # root = fv.get_widget()
    # root.show()

    # if len(args) > 0:
    #     fv.load_file(args[0])

    #fv.app.run(None)
    fv.mainloop()
    print("application finished")


if __name__ == "__main__":

    # Parse command line options
    from argparse import ArgumentParser

    argprs = ArgumentParser()

    argprs.add_argument("-r", "--render", dest="render", default='widget',
                        help="Set render type {widget|opengl}")
    log.addlogopts(argprs)

    (options, args) = argprs.parse_known_args(sys.argv[1:])

    main(options, args)



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
"""
Plugin to create an image mosaic via the collage method.

**Plugin Type: Local**

``Collage`` is a local plugin, which means it is associated with a
channel.  An instance can be opened for each channel.

**Usage**

This plugin is used to automatically create a mosaic collage in the
channel viewer using images provided by the user. The position of an image
on the collage is determined by its WCS without distortion correction.
This is meant as a quick-look tool, not a replacement for image drizzling
that takes account of image distortion, etc.

The collage only exists as a plot on the Ginga canvas.  No new single image
is actually built (if you want that, see the "Mosaic" plugin).  Some plugins
that expect to operate on single images may not work correctly with a collage.

To create a new collage, click the "New Collage" button and drag files onto
the display window (e.g. files can be dragged from the `FBrowser` plugin).
Images must have a working WCS.  The first image processed will be loaded
and its WCS will be used to orient the other tiles.
You can add new images to an existing collage simply by dragging addtional
files.

**Controls**

The "Method" control is used to choose a method for mosaicing the images in
the collage.  It has two values: 'simple' and 'warp':

- 'simple' will attempt to rotate and flip the images according to the WCS.
  It is a fast method, at the expense of accuracy.  It will not handle
  distortions near the edge of the field that should skew the image.
- 'warp' will use the WCS to completely move each pixel in the image
  according to the reference image's WCS.  This may leave empty pixels in
  the image that are filled in by sampling from surrounding pixels.
  This will be slower than the simple method, and the time increases
  linearly with the size of the images.

Check the "Collage HDUs" button to have `Collage` attempt to plot all the
image HDUs in a dragged file instead of just the first found one.

Check "Label Images" to have the plugin draw the name of each image over each
plotted tile.

If "Match bg" is checked, the background of each tile is adjusted relative
to the median of the first tile plotted (a kind of rough smoothing).

The "Num Threads" box assigns how many threads will be used from the thread
pool to load the data.  Using several threads will usually speed up loading
of many files.

**Difference from `Mosaic` plugin**

- Doesn't allocate a large array to hold all the mosaic contents
- No need to specify output FOV or worry about it
- Can be quicker to show result (depends a bit on constituent images)
- Some plugins will not work correctly with a collage, or will be slower
- Cannot save the collage as a data file (although you can use "ScreenShot")

"""
import time
import threading

from ginga.AstroImage import AstroImage
from ginga.util import dp
from ginga.util.io import io_fits
from ginga.util.mosaic import CanvasMosaicer
from ginga import GingaPlugin
from ginga.gw import Widgets


__all__ = ['Collage']


class Collage(GingaPlugin.LocalPlugin):

    def __init__(self, fv, fitsimage):
        # superclass defines some variables for us, like logger
        super(Collage, self).__init__(fv, fitsimage)

        # Load plugin preferences
        prefs = self.fv.get_preferences()
        self.settings = prefs.create_category('plugin_Collage')
        self.settings.set_defaults(annotate_images=False,
                                   annotate_color='pink',
                                   annotate_fontsize=10.0,
                                   match_bg=False,
                                   num_threads=4,
                                   collage_hdus=False,
                                   collage_method='simple',
                                   center_image=True)
        self.settings.load(onError='silent')

        self.mosaicer = CanvasMosaicer(self.logger, settings=self.settings)
        self.mosaicer.add_callback('progress', self._plot_progress_cb)
        self.mosaicer.add_callback('finished', self._plot_finished_cb)

        self.ev_intr = threading.Event()
        self.lock = threading.RLock()
        # holds processed images to be inserted into collage image
        self.images = []
        self.ingest_count = 0
        self.total_files = 0
        self.num_groups = 0
        # can set this to annotate images with a specific
        # value drawn from the FITS kwd
        self.ann_fits_kwd = None

        self.dc = self.fv.get_draw_classes()

        canvas = self.dc.DrawingCanvas()
        canvas.enable_draw(False)
        canvas.add_callback('drag-drop', self.drop_cb)
        canvas.set_surface(fitsimage)
        canvas.ui_set_active(True, viewer=fitsimage)
        self.canvas = canvas
        self.layertag = 'collage-canvas'

        # hook to allow special processing before inlining
        self.preprocess = None

        self.gui_up = False

    def build_gui(self, container):
        top = Widgets.VBox()
        top.set_border_width(4)

        vbox, sw, orientation = Widgets.get_oriented_box(container,
                                                         orientation=self.settings.get('orientation', None))
        vbox.set_border_width(4)
        vbox.set_spacing(2)

        fr = Widgets.Frame("Collage")

        captions = [
            ("New Collage", 'button', "Method:", 'label', 'method', 'combobox'),
            ("Collage HDUs", 'checkbutton', "Label images", 'checkbutton',
             "Match bg", 'checkbutton'),
            ("Num Threads:", 'label', 'Num Threads', 'llabel',
             'set_num_threads', 'entryset'),
        ]
        w, b = Widgets.build_info(captions, orientation=orientation)
        self.w.update(b)

        b.new_collage.add_callback('activated', lambda w: self.new_collage_cb())

        combobox = b.method
        options = ['simple', 'warp']
        for name in options:
            combobox.append_text(name)
        method = self.settings.get('collage_method', 'simple')
        combobox.set_text(method)
        combobox.add_callback('activated', self.set_collage_method_cb)
        combobox.set_tooltip("Choose collage method: %s" % ','.join(options))

        collage_hdus = self.settings.get('collage_hdus', False)
        b.collage_hdus.set_tooltip("Collage data HDUs in each file")
        b.collage_hdus.set_state(collage_hdus)
        b.collage_hdus.add_callback('activated', self.collage_hdus_cb)

        labelem = self.settings.get('annotate_images', False)
        b.label_images.set_state(labelem)
        b.label_images.set_tooltip("Label tiles with their names")
        b.label_images.add_callback('activated', self.annotate_cb)

        match_bg = self.settings.get('match_bg', False)
        b.match_bg.set_tooltip("Try to match background levels")
        b.match_bg.set_state(match_bg)
        b.match_bg.add_callback('activated', self.match_bg_cb)

        num_threads = self.settings.get('num_threads', 4)
        b.num_threads.set_text(str(num_threads))
        #b.set_num_threads.set_length(8)
        b.set_num_threads.set_text(str(num_threads))
        b.set_num_threads.set_tooltip("Number of threads to use for mosaicing")
        b.set_num_threads.add_callback('activated', self.set_num_threads_cb)

        fr.set_widget(w)
        vbox.add_widget(fr, stretch=0)

        vbox2 = Widgets.VBox()
        # Collage evaluation status
        hbox = Widgets.HBox()
        hbox.set_spacing(4)
        hbox.set_border_width(4)
        label = Widgets.Label()
        self.w.eval_status = label
        hbox.add_widget(self.w.eval_status, stretch=0)
        hbox.add_widget(Widgets.Label(''), stretch=1)
        vbox2.add_widget(hbox, stretch=0)

        # Collage evaluation progress bar and stop button
        hbox = Widgets.HBox()
        hbox.set_spacing(4)
        hbox.set_border_width(4)
        btn = Widgets.Button("Stop")
        btn.add_callback('activated', lambda w: self.stop_cb())
        btn.set_enabled(False)
        self.w.btn_intr_eval = btn
        hbox.add_widget(btn, stretch=0)

        self.w.eval_pgs = Widgets.ProgressBar()
        hbox.add_widget(self.w.eval_pgs, stretch=1)

        vbox2.add_widget(hbox, stretch=0)
        vbox2.add_widget(Widgets.Label(''), stretch=1)
        vbox.add_widget(vbox2, stretch=1)

        self.w.vbox = Widgets.VBox()
        vbox.add_widget(self.w.vbox, stretch=0)

        spacer = Widgets.Label('')
        vbox.add_widget(spacer, stretch=1)

        top.add_widget(sw, stretch=1)

        btns = Widgets.HBox()
        btns.set_spacing(3)

        btn = Widgets.Button("Close")
        btn.add_callback('activated', lambda w: self.close())
        btns.add_widget(btn, stretch=0)
        btn = Widgets.Button("Help")
        btn.add_callback('activated', lambda w: self.help())
        btns.add_widget(btn, stretch=0)
        btns.add_widget(Widgets.Label(''), stretch=1)
        top.add_widget(btns, stretch=0)

        container.add_widget(top, stretch=1)
        self.gui_up = True

    def set_preprocess(self, fn):
        self.preprocess = fn

    def close(self):
        self.canvas.delete_all_objects()
        self.mosaicer.reset()
        self.fv.stop_local_plugin(self.chname, str(self))
        self.gui_up = False
        return True

    def start(self):
        # insert layer if it is not already
        p_canvas = self.fitsimage.get_canvas()
        try:
            p_canvas.get_object_by_tag(self.layertag)

        except KeyError:
            # Add canvas layer
            p_canvas.add(self.canvas, tag=self.layertag)

        # image loaded in channel is used as initial reference image,
        # if there is one
        ref_image = self.fitsimage.get_image()
        if ref_image is not None:
            self.mosaicer.prepare_mosaic(ref_image)
        else:
            self.mosaicer.reset()

        self.resume()

    def stop(self):
        self.canvas.ui_set_active(False)
        p_canvas = self.fitsimage.get_canvas()
        try:
            p_canvas.delete_object_by_tag(self.layertag)
        except Exception:
            pass
        self.fitsimage.clear()
        self.fitsimage.reset_limits()
        self.mosaicer.reset()
        self.fv.show_status("")

    def pause(self):
        # comment this to NOT disable the UI for this plugin
        # when it loses focus
        #self.canvas.ui_set_active(False)
        pass

    def resume(self):
        self.canvas.ui_set_active(True, viewer=self.fitsimage)

    def new_collage_cb(self):
        self.mosaicer.reset()
        self.canvas.delete_all_objects()
        self.fitsimage.clear()
        self.fitsimage.onscreen_message("Drag new files...",
                                        delay=2.0)

    def drop_cb(self, canvas, paths, *args):
        self.logger.info("files dropped: %s" % str(paths))
        self.fv.gui_do(self.fv.error_wrap, self.collage, paths)
        return True

    def annotate_cb(self, widget, tf):
        self.settings.set(annotate_images=tf)
        self.mosaicer.annotate_images(self.canvas)

    def load_tiles(self, paths, image_loader=None, preprocess=None):
        self.fv.assert_nongui_thread()

        if image_loader is None:
            image_loader = self.fv.load_image
        if preprocess is None:
            preprocess = self.preprocess

        try:
            for url in paths:
                if self.ev_intr.is_set():
                    break
                collage_hdus = self.settings.get('collage_hdus', False)
                if collage_hdus:
                    self.logger.debug("loading hdus")
                    opener = io_fits.get_fitsloader(logger=self.logger)
                    # User wants us to collage HDUs
                    opener.open_file(url, memmap=False)
                    try:
                        for i in range(len(opener)):
                            if self.ev_intr.is_set():
                                break
                            self.logger.debug("ingesting hdu #%d" % (i))
                            try:
                                image = opener.load_idx(i)

                            except Exception as e:
                                self.logger.error(
                                    "Failed to open HDU #%d: %s" % (i, str(e)))
                                continue

                            if not isinstance(image, AstroImage):
                                self.logger.debug(
                                    "HDU #%d is not an image; skipping..." % (i))
                                continue

                            data = image.get_data()
                            if data is None:
                                # skip blank data
                                continue

                            if image.ndim != 2:
                                # skip images without 2 dimensions
                                continue

                            wd, ht = image.get_size()
                            if wd == 0 or ht == 0:
                                # or size 0 in either dimension
                                continue

                            if preprocess is not None:
                                image = preprocess(image)

                            with self.lock:
                                self.images.append(image)

                    finally:
                        opener.close()
                        opener = None

                else:
                    image = image_loader(url)

                    if preprocess is not None:
                        image = preprocess(image)

                    with self.lock:
                        self.images.append(image)

                with self.lock:
                    self.ingest_count += 1
                    self.update_progress(self.ingest_count / self.total_files)

        finally:
            with self.lock:
                self.num_groups -= 1
                if self.num_groups <= 0:
                    images, self.images = self.images, []
                    self.fv.gui_do(self.finish_collage, images)

    def finish_collage(self, images):
        self.fv.assert_gui_thread()

        self.load_time = time.time() - self.start_time
        self.logger.info("num images={}".format(len(images)))

        if self.ev_intr.is_set():
            self.update_status("collage cancelled!")
            self.end_progress()
            return

        self.w.eval_pgs.set_value(0.0)

        self.mosaicer.mosaic(self.fitsimage, images, canvas=self.canvas,
                             ev_intr=self.ev_intr)
        self.w.btn_intr_eval.set_enabled(False)

    def collage(self, paths, image_loader=None, preprocess=None,
                new_collage=False):
        """Collage images obtained by loading ``paths`` using optional
        ``image_loader`` and ``proprocess`` (post-load, pre-collage)
        functions.  A new collage is started if ``new_collage`` is True
        otherwise an existing collage is built on.
        """
        self.fv.assert_gui_thread()

        if new_collage:
            self.mosaicer.reset()
            self.canvas.delete_all_objects()
            self.fitsimage.clear()

        self.ingest_count = 0
        self.total_files = len(paths)
        if self.total_files == 0:
            return

        self.images = []
        self.ev_intr.clear()
        # Initialize progress bar
        self.update_status("Loading files...")
        self.init_progress()
        self.start_time = time.time()

        num_threads = self.settings.get('num_threads', 4)
        groups = dp.split_n(paths, num_threads)
        self.num_groups = len(groups)
        self.logger.info("num groups=%d" % (self.num_groups))

        for group in groups:
            self.fv.nongui_do(self.load_tiles, group,
                              image_loader=image_loader, preprocess=preprocess)

    def match_bg_cb(self, w, tf):
        self.settings.set(match_bg=tf)

    def collage_hdus_cb(self, w, tf):
        self.settings.set(collage_hdus=tf)

    def set_collage_method_cb(self, w, idx):
        method = w.get_text()
        self.settings.set(collage_method=method)

    def set_num_threads_cb(self, w):
        num_threads = int(w.get_text())
        self.w.num_threads.set_text(str(num_threads))
        self.settings.set(num_threads=num_threads)

    def update_status(self, text):
        if self.gui_up:
            self.fv.gui_do(self.w.eval_status.set_text, text)
            self.fv.gui_do(self.fv.update_pending)

    def _plot_progress_cb(self, mosaicer, category, pct):
        if self.gui_up:
            self.w.eval_status.set_text(category + '...')
            self.w.eval_pgs.set_value(pct)

    def _plot_finished_cb(self, mosaicer, t_sec):
        total = self.load_time + t_sec
        msg = "done. load: %.4f collage: %.4f total: %.4f sec" % (
            self.load_time, t_sec, total)
        self.update_status(msg)

    def init_progress(self):
        def _foo():
            self.w.btn_intr_eval.set_enabled(True)
            self.w.eval_pgs.set_value(0.0)
        if self.gui_up:
            self.fv.gui_do(_foo)

    def update_progress(self, pct):
        def _foo():
            self.w.eval_pgs.set_value(pct)
        if self.gui_up:
            self.fv.gui_do(_foo)

    def end_progress(self):
        if self.gui_up:
            self.fv.gui_do(self.w.btn_intr_eval.set_enabled, False)

    def stop_cb(self):
        self.ev_intr.set()

    def __str__(self):
        return 'collage'


# Append module docstring with config doc for auto insert by Sphinx.
from ginga.util.toolbox import generate_cfg_example  # noqa
if __doc__ is not None:
    __doc__ += generate_cfg_example('plugin_Collage', package='ginga')



#
# Widgets.py -- wrapped Qt widgets and convenience functions
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import os.path
import pathlib
from functools import reduce

from ginga.qtw.QtHelp import (QtGui, QtCore, QTextCursor, QIcon, QPixmap,
                              QCursor, QFont)
from ginga.qtw import QtHelp

from ginga.misc import Callback, Bunch, Settings, LineHistory
from ginga.util.paths import icondir, app_icon_path

__all__ = ['WidgetError', 'WidgetBase', 'TextEntry', 'TextEntrySet',
           'GrowingTextEdit', 'TextArea', 'Label', 'Button', 'ComboBox',
           'SpinBox', 'Slider', 'Dial', 'ScrollBar', 'CheckBox', 'ToggleButton',
           'RadioButton', 'Image', 'ProgressBar', 'StatusBar', 'TreeView',
           'ContainerBase', 'Box', 'HBox', 'VBox', 'Frame',
           'Expander', 'TabWidget', 'StackWidget', 'MDIWidget', 'ScrollArea',
           'Splitter', 'GridBox', 'ToolbarAction', 'Toolbar', 'MenuAction',
           'Menu', 'Menubar', 'TopLevelMixin', 'TopLevel', 'Application',
           'Dialog', 'SaveDialog', 'DragPackage',
           'name_mangle', 'make_widget', 'hadjust', 'build_info', 'wrap']


class WidgetError(Exception):
    """For errors thrown in this module."""
    pass


_app = None


# BASE

class WidgetBase(Callback.Callbacks):

    def __init__(self):
        super(WidgetBase, self).__init__()

        self.widget = None
        # external data can be attached here
        self.extdata = Bunch.Bunch()

    def get_widget(self):
        return self.widget

    def set_tooltip(self, text):
        self.widget.setToolTip(text)

    def get_enabled(self):
        self.widget.isEnabled()

    def set_enabled(self, tf):
        self.widget.setEnabled(tf)

    def get_size(self):
        wd, ht = self.widget.width(), self.widget.height()
        return (wd, ht)

    def get_pos(self):
        x, y = self.widget.x(), self.widget.y()
        return (x, y)

    def get_app(self):
        return _app

    def delete(self):
        w, self.widget = self.widget, None
        QtHelp.delete_widget(w)
        #w.deleteLater()

    def focus(self):
        self.widget.activateWindow()
        self.widget.setFocus()
        # self.widget.raise_()

    def resize(self, width, height):
        _wd, _ht = self.get_size()
        if width < 0:
            width = _wd
        if height < 0:
            height = _ht
        self.widget.resize(int(width), int(height))

    def show(self):
        self.widget.show()

    def hide(self):
        self.widget.hide()

    def is_visible(self):
        return self.widget.isVisible()

    def get_font(self, font_family, point_size):
        font = QtHelp.get_font(font_family, point_size)
        return font

    def cfg_expand(self, horizontal='fixed', vertical='fixed'):
        """WARNING: this call has specific effects dependent on the back
        end. It is not recommended to use it unless you cannot achieve the
        proper layout without it.
        """
        policy_dict = dict(fixed=QtGui.QSizePolicy.Fixed,
                           minimum=QtGui.QSizePolicy.Minimum,
                           maximum=QtGui.QSizePolicy.Maximum,
                           preferred=QtGui.QSizePolicy.Preferred,
                           expanding=QtGui.QSizePolicy.Expanding,
                           minimumexpanding=QtGui.QSizePolicy.MinimumExpanding,
                           ignored=QtGui.QSizePolicy.Ignored)
        h_policy = QtGui.QSizePolicy.Policy(policy_dict[horizontal])
        v_policy = QtGui.QSizePolicy.Policy(policy_dict[vertical])
        self.widget.setSizePolicy(QtGui.QSizePolicy(h_policy, v_policy))


# BASIC WIDGETS

class TextEntry(WidgetBase):
    def __init__(self, text='', editable=True):
        super(TextEntry, self).__init__()

        self.widget = QtGui.QLineEdit()
        self.widget.setText(text)
        self.widget.setReadOnly(not editable)
        self.widget.returnPressed.connect(self._cb_redirect)
        self.widget.keyPressEvent_org = self.widget.keyPressEvent
        self.widget.keyPressEvent = self._key_press_event

        self.history = LineHistory.LineHistory()

        self.enable_callback('activated')

    def _cb_redirect(self, *args):
        self.history.append(self.get_text())
        self.make_callback('activated')

    def _key_press_event(self, event):
        keycode = event.key()
        if keycode in [QtCore.Qt.Key_Up]:
            try:
                self.set_text(self.history.prev())
            except ValueError:
                pass
        elif keycode in [QtCore.Qt.Key_Down]:
            try:
                self.set_text(self.history.next())
            except ValueError:
                pass
        else:
            self.widget.keyPressEvent_org(event)

    def get_text(self):
        return self.widget.text()

    def set_text(self, text):
        self.widget.setText(text)

    def set_editable(self, tf):
        self.widget.setReadOnly(not tf)

    def set_font(self, font, size=10):
        if not isinstance(font, QFont):
            font = self.get_font(font, size)
        self.widget.setFont(font)

    def set_length(self, numchars):
        # this is only supposed to set the visible length (but Qt doesn't
        # really have a good way to do that)
        # self.widget.setMaxLength(numchars)
        pass


class TextEntrySet(WidgetBase):
    def __init__(self, text='', editable=True):
        super(TextEntrySet, self).__init__()

        self.widget = QtHelp.HBox()
        self.entry = QtGui.QLineEdit()
        self.entry.setText(text)
        self.entry.setReadOnly(not editable)
        layout = self.widget.layout()
        layout.addWidget(self.entry, stretch=1)
        self.btn = QtGui.QPushButton('Set')
        self.entry.returnPressed.connect(self._cb_redirect)
        self.btn.clicked.connect(self._cb_redirect)
        layout.addWidget(self.btn, stretch=0)

        self.enable_callback('activated')

    def _cb_redirect(self, *args):
        self.make_callback('activated')

    def get_text(self):
        return self.entry.text()

    def set_text(self, text):
        self.entry.setText(text)

    def set_editable(self, tf):
        self.entry.setReadOnly(not tf)

    def set_font(self, font, size=10):
        if not isinstance(font, QFont):
            font = self.get_font(font, size)
        self.widget.setFont(font)

    def set_length(self, numchars):
        # this is only supposed to set the visible length (but Qt doesn't
        # really have a good way to do that)
        # self.widget.setMaxLength(numchars)
        pass

    def set_enabled(self, tf):
        super(TextEntrySet, self).set_enabled(tf)
        self.entry.setEnabled(tf)


class GrowingTextEdit(QtGui.QTextEdit):

    def __init__(self, *args, **kwargs):
        super(GrowingTextEdit, self).__init__(*args, **kwargs)
        self.document().documentLayout().documentSizeChanged.connect(
            self.sizeChange)
        self.heightMin = 0
        self.heightMax = 65000

    def sizeChange(self):
        docHeight = self.document().size().height()
        # add some margin to prevent auto scrollbars
        docHeight += 20
        if self.heightMin <= docHeight <= self.heightMax:
            self.setMaximumHeight(int(docHeight))


class TextArea(WidgetBase):
    def __init__(self, wrap=False, editable=False):
        super(TextArea, self).__init__()

        # tw = QtGui.QTextEdit()
        tw = GrowingTextEdit()
        tw.setReadOnly(not editable)
        if wrap:
            tw.setLineWrapMode(QtGui.QTextEdit.WidgetWidth)
        else:
            tw.setLineWrapMode(QtGui.QTextEdit.NoWrap)
        self.widget = tw

    def append_text(self, text, autoscroll=True):
        if text.endswith('\n'):
            text = text[:-1]
        self.widget.append(text)
        if not autoscroll:
            return

        self.widget.moveCursor(QTextCursor.End)
        self.widget.moveCursor(QTextCursor.StartOfLine)
        self.widget.ensureCursorVisible()

    def get_text(self):
        return self.widget.document().toPlainText()

    def clear(self):
        self.widget.clear()

    def set_text(self, text):
        self.clear()
        self.append_text(text)

    def set_editable(self, tf):
        self.widget.setReadOnly(not tf)

    def set_limit(self, numlines):
        # self.widget.setMaximumBlockCount(numlines)
        pass

    def set_font(self, font, size=10):
        if not isinstance(font, QFont):
            font = self.get_font(font, size)
        self.widget.setCurrentFont(font)

    def set_wrap(self, tf):
        if tf:
            self.widget.setLineWrapMode(QtGui.QTextEdit.WidgetWidth)
        else:
            self.widget.setLineWrapMode(QtGui.QTextEdit.NoWrap)


class Label(WidgetBase):
    def __init__(self, text='', halign='left', style='normal', menu=None):
        super(Label, self).__init__()

        lbl = QtGui.QLabel(text)
        if halign == 'left':
            lbl.setAlignment(QtCore.Qt.AlignLeft)
        elif halign == 'center':
            lbl.setAlignment(QtCore.Qt.AlignHCenter)
        elif halign == 'right':
            lbl.setAlignment(QtCore.Qt.AlignRight)

        self.widget = lbl
        lbl.mousePressEvent = self._cb_redirect
        lbl.mouseReleaseEvent = self._cb_redirect2

        if style == 'clickable':
            lbl.setSizePolicy(QtGui.QSizePolicy.Minimum,
                              QtGui.QSizePolicy.Minimum)
            lbl.setFrameStyle(QtGui.QFrame.Box | QtGui.QFrame.Raised)

        if menu is not None:
            lbl.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
            menu_w = menu.get_widget()

            def on_context_menu(point):
                if hasattr(menu_w, 'exec_'):
                    # PySide2
                    menu_w.exec_(lbl.mapToGlobal(point))
                else:
                    menu_w.exec(lbl.mapToGlobal(point))

            lbl.customContextMenuRequested.connect(on_context_menu)

        # Enable highlighting for copying
        # lbl.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)

        self.enable_callback('activated')
        self.enable_callback('released')

    def _cb_redirect(self, event):
        buttons = event.buttons()
        if buttons & QtCore.Qt.LeftButton:
            self.make_callback('activated')

    def _cb_redirect2(self, event):
        buttons = event.buttons()
        if buttons & QtCore.Qt.LeftButton:
            self.make_callback('released')

    def get_text(self):
        return self.widget.text()

    def set_text(self, text):
        self.widget.setText(text)

    def set_font(self, font, size=10):
        if not isinstance(font, QFont):
            font = self.get_font(font, size)
        self.widget.setFont(font)

    def set_color(self, fg=None, bg=None):
        self.widget.setStyleSheet(
            "QLabel { background-color: %s; color: %s; }" % (bg, fg))

    def set_halign(self, align):
        # TODO: set horizontal alignment of text
        pass


class Button(WidgetBase):
    def __init__(self, text=''):
        super(Button, self).__init__()

        self.widget = QtGui.QPushButton(text)
        self.widget.clicked.connect(self._cb_redirect)

        self.enable_callback('activated')

    def set_text(self, text):
        self.widget.setText(text)

    def get_text(self):
        return self.widget.text()

    def _cb_redirect(self, *args):
        self.make_callback('activated')


class ComboBox(WidgetBase):
    def __init__(self, editable=False):
        super(ComboBox, self).__init__()

        self.widget = QtHelp.ComboBox()
        self.widget.setSizeAdjustPolicy(QtGui.QComboBox.AdjustToContents)
        self.widget.setEditable(editable)
        self.widget.activated.connect(self._cb_redirect)

        self.enable_callback('activated')

    def _cb_redirect(self, *args):
        idx = self.widget.currentIndex()
        self.make_callback('activated', idx)

    def insert_alpha(self, text):
        index = 0
        while True:
            itemText = self.widget.itemText(index)
            if len(itemText) == 0:
                break
            if itemText > text:
                self.widget.insertItem(index, text)
                return
            index += 1
        self.widget.addItem(text)

    def delete_alpha(self, text):
        index = self.widget.findText(text)
        self.widget.removeItem(index)

    def get_alpha(self, idx):
        return self.widget.itemText(idx)

    def clear(self):
        self.widget.clear()

    def set_text(self, text):
        self.widget.blockSignals(True)
        index = self.widget.findText(text)
        if index >= 0:
            self.set_index(index)
        else:
            self.widget.setEditText(text)
        self.widget.blockSignals(False)

    # to be deprecated someday
    show_text = set_text

    def get_text(self):
        idx = self.get_index()
        return self.get_alpha(idx)

    def append_text(self, text):
        self.widget.addItem(text)

    def set_index(self, index):
        self.widget.blockSignals(True)
        self.widget.setCurrentIndex(index)
        self.widget.blockSignals(False)

    def get_index(self):
        return self.widget.currentIndex()


class SpinBox(WidgetBase):
    def __init__(self, dtype=int):
        super(SpinBox, self).__init__()

        if dtype == float:
            w = QtGui.QDoubleSpinBox()
        else:
            w = QtGui.QSpinBox()
        w.valueChanged.connect(self._cb_redirect)
        # should values wrap around
        w.setWrapping(False)
        self.widget = w

        self.enable_callback('value-changed')

    def _cb_redirect(self, val):
        self.make_callback('value-changed', val)

    def get_value(self):
        return self.widget.value()

    def set_value(self, val):
        self.widget.blockSignals(True)
        self.widget.setValue(val)
        self.widget.blockSignals(False)

    def set_decimals(self, num):
        if hasattr(self.widget, 'setDecimals'):
            # only for QDoubleSpinBox
            self.widget.setDecimals(num)

    def set_limits(self, minval, maxval, incr_value=1):
        adj = self.widget
        adj.setRange(minval, maxval)
        adj.setSingleStep(incr_value)


class Slider(WidgetBase):
    def __init__(self, orientation='horizontal', dtype=int, track=False):
        super(Slider, self).__init__()

        # NOTE: parameter dtype is ignored for now for Qt

        if orientation == 'horizontal':
            w = QtGui.QSlider(QtCore.Qt.Horizontal)
            w.setTickPosition(QtGui.QSlider.TicksBelow)
        else:
            w = QtGui.QSlider(QtCore.Qt.Vertical)
            w.setTickPosition(QtGui.QSlider.TicksRight)
        # w.setTickPosition(QtGui.QSlider.NoTicks)
        # this controls whether the callbacks are made *as the user
        # moves the slider* or afterwards
        w.setTracking(track)
        self.widget = w
        w.valueChanged.connect(self._cb_redirect)

        self.enable_callback('value-changed')

    def _cb_redirect(self, val):
        self.make_callback('value-changed', val)

    def get_value(self):
        return self.widget.value()

    def set_value(self, val):
        self.widget.blockSignals(True)
        self.widget.setValue(val)
        self.widget.blockSignals(False)

    def set_tracking(self, tf):
        self.widget.setTracking(tf)

    def set_limits(self, minval, maxval, incr_value=1):
        adj = self.widget
        adj.setRange(minval, maxval)
        adj.setSingleStep(incr_value)


class Dial(WidgetBase):
    def __init__(self, dtype=float, wrap=False, track=False):
        super(Dial, self).__init__()

        w = QtGui.QDial()
        w.setWrapping(wrap)
        w.setNotchesVisible(True)

        self._precision = 10000
        w.setRange(0, self._precision)
        w.setSingleStep(int(self._precision / 100))
        self.dtype = dtype
        self.min_val = dtype(0)
        self.max_val = dtype(100)
        self.inc_val = dtype(1)

        # this controls whether the callbacks are made *as the user
        # moves the slider* or afterwards
        w.setTracking(track)
        self.widget = w
        w.valueChanged.connect(self._cb_redirect)

        self.enable_callback('value-changed')

    def _cb_redirect(self, val):
        val = self.get_value()
        self.make_callback('value-changed', val)

    def _cvt_value_out(self, int_val):
        pct = int_val / self._precision
        rng = self.max_val - self.min_val
        val = self.dtype(self.min_val + pct * rng)
        return val

    def _cvt_value_in(self, ext_val):
        rng = self.max_val - self.min_val
        pct = (ext_val - self.min_val) / rng
        val = int(pct * self._precision)
        return val

    def get_value(self):
        int_val = self.widget.value()
        return self._cvt_value_out(int_val)

    def set_value(self, val):
        if val < self.min_val or val > self.max_val:
            raise ValueError("Value '{}' is out of range".format(val))
        self.changed = True
        int_val = self._cvt_value_in(val)
        self.widget.blockSignals(True)
        self.widget.setValue(int_val)
        self.widget.blockSignals(False)

    def set_tracking(self, tf):
        self.widget.setTracking(tf)

    def set_limits(self, minval, maxval, incr_value=1):
        self.min_val = minval
        self.max_val = maxval
        self.inc_val = incr_value

        int_val = self._cvt_value_in(incr_value)
        self.widget.setSingleStep(int_val)


class ScrollBar(WidgetBase):
    def __init__(self, orientation='horizontal'):
        super(ScrollBar, self).__init__()

        if orientation == 'horizontal':
            self.widget = QtGui.QScrollBar(QtCore.Qt.Horizontal)
        else:
            self.widget = QtGui.QScrollBar(QtCore.Qt.Vertical)
        self.widget.valueChanged.connect(self._cb_redirect)

        self.enable_callback('activated')

    def set_value(self, value):
        int_val = int(round(value * 100.0))
        self.widget.setValue(int_val)

    def get_value(self):
        return self.widget.value() / 100.0

    def _cb_redirect(self):
        val = self.widget.value() / 100.0
        self.make_callback('activated', val)


class CheckBox(WidgetBase):
    def __init__(self, text=''):
        super(CheckBox, self).__init__()

        self.widget = QtGui.QCheckBox(text)
        self.widget.stateChanged.connect(self._cb_redirect)

        self.enable_callback('activated')

    def _cb_redirect(self, *args):
        val = self.get_state()
        self.make_callback('activated', val)

    def set_state(self, tf):
        self.widget.blockSignals(True)
        self.widget.setChecked(tf)
        self.widget.blockSignals(False)

    def get_state(self):
        val = self.widget.checkState()
        # returns 0 (unchecked) or 2 (checked)
        return (val == QtCore.Qt.CheckState.Checked)


class ToggleButton(WidgetBase):
    def __init__(self, text=''):
        super(ToggleButton, self).__init__()

        self.widget = QtGui.QPushButton(text)
        self.widget.setCheckable(True)
        self.widget.clicked.connect(self._cb_redirect)

        self.enable_callback('activated')

    def _cb_redirect(self, val):
        self.make_callback('activated', val)

    def set_state(self, tf):
        self.widget.blockSignals(True)
        self.widget.setChecked(tf)
        self.widget.blockSignals(False)

    def get_state(self):
        return self.widget.isChecked()


class RadioButton(WidgetBase):
    def __init__(self, text='', group=None):
        super(RadioButton, self).__init__()

        self.widget = QtGui.QRadioButton(text)
        self.widget.toggled.connect(self._cb_redirect)

        self.enable_callback('activated')

    def _cb_redirect(self, val):
        self.make_callback('activated', val)

    def set_state(self, tf):
        if self.widget.isChecked() != tf:
            # toggled only fires when the value is toggled
            self.widget.blockSignals(True)
            self.widget.setChecked(tf)
            self.widget.blockSignals(False)

    def get_state(self):
        return self.widget.isChecked()


class Image(WidgetBase):
    def __init__(self, native_image=None, style='normal', menu=None):
        super(Image, self).__init__()

        lbl = QtGui.QLabel()
        self.widget = lbl
        if native_image is not None:
            self._set_image(native_image)

        lbl.mousePressEvent = self._cb_redirect

        if style == 'clickable':
            lbl.setSizePolicy(QtGui.QSizePolicy.Minimum,
                              QtGui.QSizePolicy.Minimum)
            # lbl.setFrameStyle(QtGui.QFrame.Box | QtGui.QFrame.Raised)

        if menu is not None:
            lbl.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
            menu_w = menu.get_widget()

            def on_context_menu(point):
                if hasattr(menu_w, 'exec_'):
                    # PySide2
                    menu_w.exec_(lbl.mapToGlobal(point))
                else:
                    menu_w.exec(lbl.mapToGlobal(point))

            lbl.customContextMenuRequested.connect(on_context_menu)

        self.enable_callback('activated')

    def _cb_redirect(self, event):
        buttons = event.buttons()
        if buttons & QtCore.Qt.LeftButton:
            self.make_callback('activated')

    def _set_image(self, native_image):
        pixmap = QPixmap.fromImage(native_image)
        self.widget.setPixmap(pixmap)

    def load_file(self, img_path, format=None):
        pixmap = QPixmap()
        pixmap.load(img_path, format=format)
        self.widget.setPixmap(pixmap)


class ProgressBar(WidgetBase):
    def __init__(self):
        super(ProgressBar, self).__init__()

        w = QtGui.QProgressBar()
        w.setRange(0, 100)
        w.setTextVisible(True)
        self.widget = w

    def set_value(self, pct):
        self.widget.setValue(int(pct * 100.0))


class StatusBar(WidgetBase):
    def __init__(self):
        super(StatusBar, self).__init__()

        sbar = QtGui.QStatusBar()
        sbar.setSizeGripEnabled(True)
        self.widget = sbar

    def clear_message(self):
        self.widget.showMessage('', 0)

    def set_message(self, msg_str, duration=10.0):
        # remove message in about `duration` seconds
        self.widget.showMessage(msg_str, int(duration * 1000))


class TreeWidgetItem(QtGui.QTreeWidgetItem):
    """A hack to subclass QTreeWidgetItem to enable sorting by numbers
    in a field.
    """
    def __init__(self, *args, **kwargs):
        QtGui.QTreeWidgetItem.__init__(self, *args, **kwargs)

    def __lt__(self, otherItem):
        column = self.treeWidget().sortColumn()
        try:
            return float(self.text(column)) < float(otherItem.text(column))

        except ValueError:
            return self.text(column) < otherItem.text(column)


class TreeView(WidgetBase):
    def __init__(self, auto_expand=False, sortable=False,
                 selection='single', use_alt_row_color=False,
                 dragable=False):
        super(TreeView, self).__init__()

        self.auto_expand = auto_expand
        self.sortable = sortable
        self.dragable = dragable
        self.selection = selection
        self.levels = 1
        self.leaf_key = None
        self.leaf_idx = 0
        self.columns = []
        self.datakeys = []
        # shadow index
        self.shadow = {}

        tv = QtGui.QTreeWidget()
        self.widget = tv
        tv.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
        if selection == 'multiple':
            tv.setSelectionMode(QtGui.QAbstractItemView.ExtendedSelection)
        tv.setAlternatingRowColors(use_alt_row_color)
        tv.itemDoubleClicked.connect(self._cb_redirect)
        tv.itemSelectionChanged.connect(self._selection_cb)
        if self.dragable:
            tv.setDragEnabled(True)
            tv.startDrag = self._start_drag

        for cbname in ('selected', 'activated', 'drag-start'):
            self.enable_callback(cbname)

    def setup_table(self, columns, levels, leaf_key):
        self.clear()

        self.columns = columns
        self.levels = levels
        self.leaf_key = leaf_key
        treeview = self.widget
        treeview.setColumnCount(len(columns))
        treeview.setSortingEnabled(self.sortable)

        # speeds things up a bit
        treeview.setUniformRowHeights(True)

        # create the column headers
        if not isinstance(columns[0], str):
            # columns specifies a mapping
            headers = [col[0] for col in columns]
            datakeys = [col[1] for col in columns]
        else:
            headers = datakeys = columns

        self.datakeys = datakeys
        self.leaf_idx = datakeys.index(self.leaf_key)

        if self.sortable:
            # Sort increasing by default
            treeview.sortByColumn(self.leaf_idx, QtCore.Qt.AscendingOrder)

        treeview.setHeaderLabels(headers)

    def set_tree(self, tree_dict):
        self.clear()
        self.add_tree(tree_dict)

    def add_tree(self, tree_dict):

        if self.sortable:
            self.widget.setSortingEnabled(False)

        for key in tree_dict:
            self._add_subtree(1, self.shadow,
                              self.widget, key, tree_dict[key])

        if self.sortable:
            self.widget.setSortingEnabled(True)

        # User wants auto expand?
        if self.auto_expand:
            self.widget.expandAll()

    def _add_subtree(self, level, shadow, parent_item, key, node):

        if level >= self.levels:
            # leaf node
            values = ['' if _key == 'icon' else str(node[_key])
                      for _key in self.datakeys]
            try:
                bnch = shadow[key]
                item = bnch.item
                # TODO: update leaf item

            except KeyError:
                # new item
                item = TreeWidgetItem(parent_item, values)
                if level == 1:
                    parent_item.addTopLevelItem(item)
                else:
                    parent_item.addChild(item)

                shadow[key] = Bunch.Bunch(node=node, item=item, terminal=True)

                # hack for adding an image to a table
                # TODO: add types for columns
                if 'icon' in node:
                    i = self.datakeys.index('icon')
                    item.setIcon(i, node['icon'])

                # mark cell as non-editable
                item.setFlags(item.flags() & ~QtCore.Qt.ItemIsEditable)

        else:
            try:
                # node already exists
                bnch = shadow[key]
                item = bnch.item
                d = bnch.node

            except KeyError:
                # new node
                item = TreeWidgetItem(parent_item, [str(key)])
                if level == 1:
                    parent_item.addTopLevelItem(item)
                else:
                    parent_item.addChild(item)
                d = {}
                shadow[key] = Bunch.Bunch(node=d, item=item, terminal=False)

            # recurse for non-leaf interior node
            for key in node:
                self._add_subtree(level + 1, d, item, key, node[key])

    def _selection_cb(self):
        res_dict = self.get_selected()
        self.make_callback('selected', res_dict)

    def _cb_redirect(self, item):
        res_dict = {}
        self._get_item(res_dict, item)
        self.make_callback('activated', res_dict)

    def _get_path(self, item):
        if item is None:
            return []

        if item.childCount() == 0:
            path_rest = self._get_path(item.parent())
            myname = item.text(self.leaf_idx)
            path_rest.append(myname)
            return path_rest

        myname = item.text(0)
        path_rest = self._get_path(item.parent())
        path_rest.append(myname)
        return path_rest

    def _get_item(self, res_dict, item):
        # from the QTreeViewItem `item`, return the item via a path
        # in the dictionary `res_dict`
        path = self._get_path(item)
        d, s = res_dict, self.shadow
        for name in path[:-1]:
            d = d.setdefault(name, {})
            s = s[name].node

        dst_key = path[-1]
        try:
            d[dst_key] = s[dst_key].node
        except KeyError:
            d[dst_key] = None

    def get_selected(self):
        items = list(self.widget.selectedItems())
        res_dict = {}
        for item in items:
            if item.childCount() > 0:
                # only leaf nodes can be selected
                continue
            self._get_item(res_dict, item)
        return res_dict

    def clear(self):
        self.widget.clear()
        self.shadow = {}

    def clear_selection(self):
        self.widget.blockSignals(True)
        try:
            self.widget.clearSelection()
        finally:
            self.widget.blockSignals(False)

    def _path_to_item(self, path):
        s = self.shadow
        for name in path[:-1]:
            s = s[name].node
        item = s[path[-1]].item
        return item

    def select_path(self, path, state=True):
        item = self._path_to_item(path)
        self.widget.blockSignals(True)
        try:
            item.setSelected(state)
        finally:
            self.widget.blockSignals(False)

    def get_selected_paths(self):
        items = list(self.widget.selectedItems())
        return [self._get_path(item) for item in items]

    def select_paths(self, paths, state=True):
        self.widget.blockSignals(True)
        try:
            for path in paths:
                item = self._path_to_item(path)
                item.setSelected(state)
        finally:
            self.widget.blockSignals(False)

    def highlight_path(self, path, onoff, font_color='green'):
        item = self._path_to_item(path)

        # A little painfully inefficient, can we do better than this?
        font = QtHelp.QFont()
        if not onoff:
            color = QtHelp.QColor('black')
        else:
            font.setBold(True)
            color = QtHelp.QColor(font_color)
        brush = QtHelp.QBrush(color)

        for i in range(item.columnCount()):
            item.setForeground(i, brush)
            item.setFont(i, font)

    def scroll_to_path(self, path):
        item = self._path_to_item(path)
        midx = self.widget.indexFromItem(item, 0)
        self.widget.scrollTo(midx, QtGui.QAbstractItemView.PositionAtCenter)

    def scroll_to_end(self):
        model = self.widget.model()
        midx = model.index(model.rowCount() - 1, 0)
        self.widget.scrollTo(midx, QtGui.QAbstractItemView.PositionAtBottom)

    def sort_on_column(self, i):
        self.widget.sortByColumn(i, QtCore.Qt.AscendingOrder)

    def set_column_width(self, i, width):
        self.widget.setColumnWidth(i, width)

    def set_column_widths(self, lwidths):
        for i, width in enumerate(lwidths):
            if width is not None:
                self.set_column_width(i, width)

    def set_optimal_column_widths(self):
        for i in range(len(self.columns)):
            self.widget.resizeColumnToContents(i)

    def get_column_widths(self):
        return [self.widget.columnWidth(i) for i in range(len(self.columns))]

    def _start_drag(self, event):
        res_dict = self.get_selected()
        drag_pkg = DragPackage(self.widget)
        self.make_callback('drag-start', drag_pkg, res_dict)
        drag_pkg.start_drag()


# CONTAINERS

class ContainerBase(WidgetBase):
    def __init__(self):
        super(ContainerBase, self).__init__()
        self.children = []

        for name in ['widget-added', 'widget-removed']:
            self.enable_callback(name)

    def add_ref(self, ref):
        # TODO: should this be a weakref?
        self.children.append(ref)

    def _remove(self, childw, delete=False):
        layout = self.widget.layout()
        if layout is not None:
            layout.removeWidget(childw)

        childw.setParent(None)
        if delete:
            childw.deleteLater()

    def remove(self, child, delete=False):
        if child not in self.children:
            raise ValueError("Widget is not a child of this container")
        self.children.remove(child)

        self._remove(child.get_widget(), delete=delete)
        self.make_callback('widget-removed', child)

    def remove_all(self, delete=False):
        for w in list(self.children):
            self.remove(w, delete=delete)

    def get_children(self):
        return self.children

    def num_children(self):
        return len(self.children)

    def _get_native_children(self):
        return [child.get_widget() for child in self.children]

    def _get_native_index(self, nchild):
        l = self._get_native_children()
        try:
            return l.index(nchild)
        except (IndexError, ValueError) as e:
            return -1

    def _native_to_child(self, nchild):
        idx = self._get_native_index(nchild)
        if idx < 0:
            return None
        return self.children[idx]

    def set_margins(self, left, right, top, bottom):
        layout = self.widget.layout()
        layout.setContentsMargins(left, right, top, bottom)

    def set_border_width(self, pix):
        layout = self.widget.layout()
        layout.setContentsMargins(pix, pix, pix, pix)


class Box(ContainerBase):
    def __init__(self, orientation='horizontal'):
        super(Box, self).__init__()

        self.widget = QtGui.QWidget()
        self.orientation = orientation
        if orientation == 'horizontal':
            self.layout = QtGui.QHBoxLayout()
        else:
            self.layout = QtGui.QVBoxLayout()

        # because of ridiculous defaults
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.widget.setLayout(self.layout)

    def insert_widget(self, idx, child, stretch=0):
        self.add_ref(child)
        child_w = child.get_widget()
        self.layout.insertWidget(idx, child_w, stretch=stretch)
        self.make_callback('widget-added', child)

    def add_widget(self, child, stretch=0):
        self.add_ref(child)
        child_w = child.get_widget()
        self.layout.addWidget(child_w, stretch=stretch)
        self.make_callback('widget-added', child)

    def set_spacing(self, val):
        self.layout.setSpacing(val)


class HBox(Box):
    def __init__(self):
        super(HBox, self).__init__(orientation='horizontal')


class VBox(Box):
    def __init__(self):
        super(VBox, self).__init__(orientation='vertical')


class Frame(ContainerBase):
    def __init__(self, title=None):
        super(Frame, self).__init__()

        self.widget = QtGui.QFrame()
        self.widget.setFrameStyle(QtGui.QFrame.Box | QtGui.QFrame.Raised)
        vbox = QtGui.QVBoxLayout()
        self.layout = vbox
        # because of ridiculous defaults
        vbox.setContentsMargins(2, 2, 2, 2)
        self.widget.setLayout(vbox)
        if title:
            lbl = QtGui.QLabel(title)
            lbl.setAlignment(QtCore.Qt.AlignHCenter)
            # lbl.setAlignment(QtCore.Qt.AlignLeft)
            vbox.addWidget(lbl, stretch=0)
            self.label = lbl
        else:
            self.label = None

    def set_widget(self, child, stretch=1):
        self.remove_all()
        self.add_ref(child)
        self.widget.layout().addWidget(child.get_widget(), stretch=stretch)

    def set_text(self, text):
        if self.label is not None:
            self.label.setText(text)


# Qt custom expander widget
# See http://stackoverflow.com/questions/10364589/equivalent-of-gtks-expander-in-pyqt4  # noqa
#
class Expander(ContainerBase):
    r_arrow = None
    d_arrow = None

    # Note: add 'text-align: left;' if you want left adjusted labels
    widget_style = """
    QPushButton { margin: 1px,1px,1px,1px; padding: 0px;
                  border-width: 0px; border-style: solid; }
    """

    def __init__(self, title='', notoggle=False):
        super(Expander, self).__init__()

        # Qt doesn't seem to like it (segfault) if we actually construct
        # these icons in the class variable declarations
        if Expander.r_arrow is None:
            Expander.r_arrow = QtHelp.get_icon(
                os.path.join(icondir, 'triangle-right.svg'), size=(12, 12))
        if Expander.d_arrow is None:
            Expander.d_arrow = QtHelp.get_icon(
                os.path.join(icondir, 'triangle-down.svg'), size=(12, 12))

        self.widget = QtGui.QWidget()
        vbox = QtGui.QVBoxLayout()
        vbox.setContentsMargins(0, 0, 0, 0)
        vbox.setSpacing(0)
        self.layout = vbox

        self.toggle = None
        if not notoggle:
            toggle = ToggleButton(title)
            self.toggle = toggle
            toggle_w = toggle.get_widget()
            toggle_w.setIcon(Expander.r_arrow)
            toggle_w.setStyleSheet(Expander.widget_style)
            toggle.add_callback('activated', self._toggle_widget)
            vbox.addWidget(toggle.get_widget(), stretch=0)

        self.widget.setLayout(vbox)

        for name in ('opened', 'closed'):
            self.enable_callback(name)

    def set_widget(self, child, stretch=1):
        self.remove_all()
        self.add_ref(child)
        child_w = child.get_widget()
        self.widget.layout().addWidget(child_w, stretch=stretch)
        child_w.setVisible(False)

    def expand(self, tf):
        child = self.get_children()[0]
        child_w = child.get_widget()
        if tf:
            if child_w.isVisible():
                # child already open
                return
            if self.toggle is not None:
                self.toggle.get_widget().setIcon(Expander.d_arrow)
            child_w.setVisible(True)
            self.make_callback('opened')
        else:
            if not child_w.isVisible():
                # child already closed
                return
            if self.toggle is not None:
                self.toggle.get_widget().setIcon(Expander.r_arrow)
            child_w.setVisible(False)
            self.make_callback('closed')

    def _toggle_widget(self, w, tf):
        self.expand(tf)


class TabWidget(ContainerBase):
    def __init__(self, tabpos='top', reorderable=False, detachable=False,
                 group=0):
        super(TabWidget, self).__init__()

        self.reorderable = reorderable
        self.detachable = detachable

        w = QtGui.QTabWidget()
        w.currentChanged.connect(self._cb_redirect)
        w.tabCloseRequested.connect(self._tab_close)
        w.setUsesScrollButtons(True)
        # w.setTabsClosable(True)
        if self.reorderable:
            w.setMovable(True)
        # w.tabInserted = self._tab_insert_cb
        # w.tabRemoved = self._tab_remove_cb
        self.widget = w
        self.set_tab_position(tabpos)

        for name in ('page-switch', 'page-close', 'page-move', 'page-detach'):
            self.enable_callback(name)

    def set_tab_position(self, tabpos):
        w = self.widget
        if tabpos == 'top':
            w.setTabPosition(QtGui.QTabWidget.North)
        elif tabpos == 'bottom':
            w.setTabPosition(QtGui.QTabWidget.South)
        elif tabpos == 'left':
            w.setTabPosition(QtGui.QTabWidget.West)
        elif tabpos == 'right':
            w.setTabPosition(QtGui.QTabWidget.East)

    def _cb_redirect(self, index):
        # get new index, because passed index can be out of date
        index = self.get_index()
        child = self.index_to_widget(index)
        if child is not None:
            self.make_callback('page-switch', child)

    def _tab_close(self, index):
        child = self.index_to_widget(index)
        self.make_callback('page-close', child)

    def add_widget(self, child, title=''):
        self.add_ref(child)
        child_w = child.get_widget()
        self.widget.addTab(child_w, title)
        # attach title to child
        child.extdata.tab_title = title
        self.make_callback('widget-added', child)

    def _remove(self, nchild, delete=False):
        idx = self.widget.indexOf(nchild)
        self.widget.removeTab(idx)

        nchild.setParent(None)
        if delete:
            nchild.deleteLater()

    def get_index(self):
        return self.widget.currentIndex()

    def set_index(self, idx):
        self.widget.setCurrentIndex(idx)
        # child = self.index_to_widget(idx)
        # child.focus()

    def index_of(self, child):
        return self.widget.indexOf(child.get_widget())

    def index_to_widget(self, idx):
        """Returns child corresponding to `idx`"""
        nchild = self.widget.widget(idx)
        if nchild is None:
            return nchild
        return self._native_to_child(nchild)

    def highlight_tab(self, idx, tf):
        tabbar = self.widget.tabBar()
        if not tf:
            color = QtHelp.QColor('black')
        else:
            color = QtHelp.QColor('green')
        tabbar.setTabTextColor(idx, color)


class StackWidget(ContainerBase):
    def __init__(self):
        super(StackWidget, self).__init__()

        self.widget = QtGui.QStackedWidget()

        for name in ['page-switch']:
            self.enable_callback(name)

    def add_widget(self, child, title=''):
        self.add_ref(child)
        child_w = child.get_widget()
        self.widget.addWidget(child_w)
        # attach title to child
        child.extdata.tab_title = title
        self.make_callback('widget-added', child)

    def get_index(self):
        return self.widget.currentIndex()

    def set_index(self, idx):
        _idx = self.widget.currentIndex()
        self.widget.setCurrentIndex(idx)

        child = self.index_to_widget(idx)
        if _idx != idx:
            self.make_callback('page-switch', child)
        # child.focus()

    def index_of(self, child):
        return self.widget.indexOf(child.get_widget())

    def index_to_widget(self, idx):
        nchild = self.widget.widget(idx)
        return self._native_to_child(nchild)


class MDIWidget(ContainerBase):
    def __init__(self, tabpos='top', mode='mdi'):
        super(MDIWidget, self).__init__()

        w = QtGui.QMdiArea()
        w.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        w.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        # See note below in add_widget()
        #w.subWindowActivated.connect(self._cb_redirect)
        # disable wheel event scrolling the space
        w.wheelEvent = self.wheelEvent

        # w.setSizePolicy(QtGui.QSizePolicy(QtGui.QSizePolicy.Expanding,
        #                                   QtGui.QSizePolicy.Expanding))
        w.setTabsClosable(True)
        w.setTabsMovable(False)
        self.widget = w
        self.true_mdi = True
        self.cur_index = -1

        for name in ('page-switch', 'page-close'):
            self.enable_callback(name)

        self.set_tab_position(tabpos)
        self.set_mode(mode)

    def set_tab_position(self, tabpos):
        w = self.widget
        if tabpos == 'top':
            w.setTabPosition(QtGui.QTabWidget.North)
        elif tabpos == 'bottom':
            w.setTabPosition(QtGui.QTabWidget.South)
        elif tabpos == 'left':
            w.setTabPosition(QtGui.QTabWidget.West)
        elif tabpos == 'right':
            w.setTabPosition(QtGui.QTabWidget.East)

    def get_mode(self):
        if self.widget.viewMode() == QtGui.QMdiArea.TabbedView:
            return 'tabs'
        return 'mdi'

    def set_mode(self, mode):
        mode = mode.lower()
        if mode == 'tabs':
            self.widget.setViewMode(QtGui.QMdiArea.TabbedView)
        elif mode == 'mdi':
            self.widget.setViewMode(QtGui.QMdiArea.SubWindowView)
        else:
            raise ValueError("Don't understand mode='%s'" % (mode))

    def _cb_redirect(self, subwin):
        if subwin is not None:
            nchild = subwin.widget()
            child = self._native_to_child(nchild)
            try:
                self.cur_index = self.children.index(child)
            except Exception:
                self.cur_index = -1
            self.make_callback('page-switch', child)

    def _child_close(self, subwin, child):
        self.make_callback('page-close', child)

    def add_widget(self, child, title=''):
        self.add_ref(child)
        subwin = MDIWindow(self, child, title=title)
        subwin.add_callback('close', self._child_close, child)

        self.make_callback('widget-added', child)
        return subwin

    def _remove(self, nchild, delete=False):
        subwins = list(self.widget.subWindowList())
        l = [sw.widget() for sw in subwins]
        try:
            idx = l.index(nchild)
            subwin = subwins[idx]
        except (IndexError, ValueError) as e:
            subwin = None

        if subwin is not None:
            self.widget.removeSubWindow(subwin)
            subwin.deleteLater()

        nchild.setParent(None)
        if delete:
            nchild.deleteLater()

    def get_index(self):
        subwin = self.widget.activeSubWindow()
        if subwin is not None:
            return self._get_native_index(subwin.widget())
        return self.cur_index

    def _get_subwin(self, widget):
        for subwin in list(self.widget.subWindowList()):
            if subwin.widget() == widget:
                return subwin
        return None

    def set_index(self, idx):
        if 0 <= idx < len(self.children):
            child = self.children[idx]
            subwin = self._get_subwin(child.widget)
            if subwin is not None:
                self.widget.setActiveSubWindow(subwin)

    def index_of(self, child):
        nchild = child.get_widget()
        return self._get_native_index(nchild)

    def index_to_widget(self, idx):
        if 0 <= idx < len(self.children):
            return self.children[idx]
        return None

    def tile_panes(self):
        self.widget.tileSubWindows()

    def cascade_panes(self):
        self.widget.cascadeSubWindows()

    def use_tabs(self, tf):
        if tf:
            self.widget.setViewMode(QtGui.QMdiArea.TabbedView)
        else:
            self.widget.setViewMode(QtGui.QMdiArea.SubWindowView)

    def wheelEvent(self, event):
        event.accept()


class ScrollArea(ContainerBase):
    def __init__(self):
        super(ScrollArea, self).__init__()

        self.widget = QtGui.QScrollArea()
        self.widget.setWidgetResizable(True)
        self.widget._resizeEvent = self.widget.resizeEvent
        self.widget.resizeEvent = self._resize_cb

        self.enable_callback('configure')

    def _resize_cb(self, event):
        self.widget._resizeEvent(event)

        rect = self.widget.geometry()
        x1, y1, x2, y2 = rect.getCoords()
        width = x2 - x1
        height = y2 - y1
        self.make_callback('configure', width, height)

    def set_widget(self, child):
        self.add_ref(child)
        self.widget.setWidget(child.get_widget())

    def scroll_to_end(self, vertical=True, horizontal=False):
        area = self.widget
        if vertical:
            area.verticalScrollBar().setValue(
                area.verticalScrollBar().maximum())
        if horizontal:
            area.horizontalScrollBar().setValue(
                area.horizontalScrollBar().maximum())


class Splitter(ContainerBase):
    def __init__(self, orientation='horizontal', thumb_px=8):
        super(Splitter, self).__init__()

        self.thumb_px = thumb_px

        w = QtGui.QSplitter()
        self.orientation = orientation
        # NOTE: need to style splitter due to lack of any visual
        # indicator on Linux and Windows
        if self.orientation == 'horizontal':
            w.setOrientation(QtCore.Qt.Horizontal)
            if thumb_px is not None:
                iconfile = pathlib.Path(icondir) / 'vdots.png'
                w.setStyleSheet(
                    """
                    QSplitter::handle { width: %spx; height: %spx;
                                        image: url(%s); }
                    """ % (self.thumb_px, self.thumb_px, iconfile))
        else:
            w.setOrientation(QtCore.Qt.Vertical)
            if thumb_px is not None:
                iconfile = pathlib.Path(icondir) / 'hdots.png'
                w.setStyleSheet(
                    """
                    QSplitter::handle { height: %spx; image: url(%s); }
                    """ % (self.thumb_px, iconfile))
        self.widget = w
        w.setChildrenCollapsible(True)

    def add_widget(self, child):
        self.add_ref(child)
        child_w = child.get_widget()
        self.widget.addWidget(child_w)
        self.make_callback('widget-added', child)

    def get_sizes(self):
        return list(self.widget.sizes())

    def set_sizes(self, sizes):
        return self.widget.setSizes(sizes)


class GridBox(ContainerBase):
    def __init__(self, rows=1, columns=1):
        super(GridBox, self).__init__()

        w = QtGui.QWidget()
        layout = QtGui.QGridLayout()
        w.setLayout(layout)
        self.widget = w
        self.tbl = {}

    def resize_grid(self, rows, columns):
        pass

    def set_row_spacing(self, val):
        self.widget.layout().setVerticalSpacing(val)

    def set_column_spacing(self, val):
        self.widget.layout().setHorizontalSpacing(val)

    def set_spacing(self, val):
        self.set_row_spacing(val)
        self.set_column_spacing(val)

    def get_row_column_count(self):
        num_rows = self.widget.layout().rowCount()
        num_cols = self.widget.layout().columnCount()
        return num_rows, num_cols

    def add_widget(self, child, row, col, stretch=0):
        key = (row, col)
        if key in self.tbl:
            # take care of case where we are overwriting a child
            old_child = self.tbl[key]
            old_child.hide()
            self.remove(old_child)
        self.tbl[key] = child
        self.add_ref(child)

        w = child.get_widget()
        self.widget.layout().addWidget(w, row, col)
        self.make_callback('widget-added', child)

    def remove(self, child, delete=False):
        super().remove(child, delete=delete)

        # need to delete the child from self.tbl
        children = list(self.tbl.values())
        if child in children:
            keys = list(self.tbl.keys())
            idx = children.index(child)
            key = keys[idx]
            del self.tbl[key]

    def get_widget_at_cell(self, row, col):
        return self.tbl[(row, col)]

    def insert_row(self, index, widgets=None):
        num_rows, num_cols = self.get_row_column_count()

        if widgets is not None:
            if len(widgets) != num_cols:
                raise ValueError("Number of widgets ({}) != number of columns ({})".format(len(widgets), num_cols))

        # handle case where user inserts row before the end of the gridbox
        if index < num_rows:
            # shift key/value pairs down to make the row empty at index
            for i in range(num_rows - 1, index - 1, -1):
                for j in range(num_cols):
                    key = (i, j)
                    if key in self.tbl:
                        child = self.tbl.pop(key)
                        self.tbl[(i + 1, j)] = child
                        # move actual widget down in QGridLayout
                        w = child.get_widget()
                        self._remove(w)
                        self.widget.layout().addWidget(w, i + 1, j)

        if widgets is not None:
            for j in range(num_cols):
                child = widgets[j]
                self.add_widget(child, index, j)

    def append_row(self, widgets):
        num_rows, num_cols = self.get_row_column_count()
        return self.insert_row(num_rows, widgets)

    def delete_row(self, index):
        num_rows, num_cols = self.get_row_column_count()
        if index < 0 or index >= num_rows:
            raise ValueError("Index ({}) out of bounds ({})".format(index, num_rows))

        # remove widgets in row to be deleted from table
        for j in range(num_cols):
            key = (index, j)
            if key in self.tbl:
                child = self.tbl.pop(key)
                child.hide()
                self.remove(child)

        if index < num_rows - 1:
            # if not removing very last row, shift key/value pairs up
            for i in range(index + 1, num_rows):
                for j in range(num_cols):
                    key = (i, j)
                    if key in self.tbl:
                        child = self.tbl.pop(key)
                        self.tbl[(i - 1, j)] = child
                        # move actual widget up in QGridLayout
                        w = child.get_widget()
                        self._remove(w)
                        self.widget.layout().addWidget(w, i - 1, j)


class ToolbarAction(WidgetBase):
    def __init__(self):
        super(ToolbarAction, self).__init__()

        self.widget = None
        self.enable_callback('activated')

    def _cb_redirect(self, *args):
        if self.widget.isCheckable():
            tf = self.widget.isChecked()
            self.make_callback('activated', tf)
        else:
            self.make_callback('activated')

    def set_state(self, tf):
        self.widget.setChecked(tf)

    def get_state(self):
        return self.widget.isChecked()


class Toolbar(ContainerBase):
    def __init__(self, orientation='horizontal'):
        super(Toolbar, self).__init__()

        w = QtGui.QToolBar()
        if orientation == 'horizontal':
            w.setOrientation(QtCore.Qt.Horizontal)
        else:
            w.setOrientation(QtCore.Qt.Vertical)
        self.widget = w
        self._menu = None
        self.widget.setStyleSheet(
            """
            QToolBar { padding: 0; spacing: 1; }\n
            QToolBar QToolButton { padding: 2; margin: 0; }\n
            """)

    def add_action(self, text, toggle=False, iconpath=None, iconsize=None):
        child = ToolbarAction()
        if iconpath:
            wd, ht = 24, 24
            if iconsize is not None:
                wd, ht = iconsize
            iconw = QtHelp.get_icon(iconpath, size=(wd, ht))
            action = self.widget.addAction(iconw, text,
                                           child._cb_redirect)
            if text is not None and len(text) > 0:
                self.widget.setToolButtonStyle(QtCore.Qt.ToolButtonTextUnderIcon)
        else:
            action = self.widget.addAction(text, child._cb_redirect)
        action.setCheckable(toggle)
        child.widget = action
        self.add_ref(child)
        return child

    def add_widget(self, child):
        self.add_ref(child)
        w = child.get_widget()
        # in toolbars, generally don't want widgets to take up any more
        # space than necessary
        w.setSizePolicy(QtGui.QSizePolicy.Fixed,
                        QtGui.QSizePolicy.Fixed)
        self.widget.addWidget(w)
        self.make_callback('widget-added', child)

    def add_menu(self, text, menu=None, mtype='tool'):
        if menu is None:
            menu = Menu()
        if mtype == 'tool':
            child = self.add_action(text)
        else:
            child = Label(text, style='clickable', menu=menu)
            self.add_widget(child)
            child.add_callback('released', lambda w: menu.hide())

        child.add_callback('activated', lambda w: menu.popup())
        return menu

    def add_separator(self):
        self.widget.addSeparator()


class MenuAction(WidgetBase):
    def __init__(self, text=None, checkable=False):
        super(MenuAction, self).__init__()

        self.widget = None
        self.text = text
        self.checkable = checkable
        self.enable_callback('activated')

    def set_state(self, tf):
        if not self.checkable:
            raise ValueError("Not a checkable menu item")
        self.widget.setChecked(tf)

    def get_state(self):
        return self.widget.isChecked()

    def _cb_redirect(self, *args):
        if self.widget.isCheckable():
            tf = self.widget.isChecked()
            self.make_callback('activated', tf)
        else:
            self.make_callback('activated')


class Menu(ContainerBase):
    def __init__(self):
        super(Menu, self).__init__()

        # NOTE: this get's overwritten if created from Menubar
        self.widget = QtGui.QMenu()
        if hasattr(self.widget, 'setToolTipsVisible'):
            # only for QT >= 5.1
            self.widget.setToolTipsVisible(True)
        self.menus = Bunch.Bunch(caseless=True)

    def add_widget(self, child):
        w = self.widget.addAction(child.text, lambda: child._cb_redirect())
        if child.checkable:
            w.setCheckable(True)
        child.widget = w
        self.add_ref(child)
        self.make_callback('widget-added', child)

    def add_name(self, name, checkable=False):
        child = MenuAction(text=name, checkable=checkable)
        self.add_widget(child)
        return child

    def add_menu(self, name):
        menu_w = self.widget.addMenu(name)
        child = Menu()
        child.widget = menu_w
        self.add_ref(child)
        self.menus[name] = child
        return child

    def add_separator(self):
        self.widget.addSeparator()

    def popup(self, widget=None):
        if widget is not None:
            w = widget.get_widget()
            if w.isEnabled():
                if hasattr(self.widget, 'exec_'):
                    # PySide2
                    self.widget.exec_(w.mapToGlobal(QtCore.QPoint(0, 0)))
                else:
                    self.widget.exec(w.mapToGlobal(QtCore.QPoint(0, 0)))
        else:
            if self.widget.isEnabled():
                if hasattr(self.widget, 'exec_'):
                    # PySide2
                    self.widget.exec_(QCursor.pos())
                else:
                    self.widget.exec(QCursor.pos())

    def get_menu(self, name):
        return self.menus[name]


class Menubar(ContainerBase):
    def __init__(self):
        super(Menubar, self).__init__()

        self.widget = QtGui.QMenuBar()
        if hasattr(self.widget, 'setNativeMenuBar'):
            self.widget.setNativeMenuBar(True)
        self.menus = Bunch.Bunch(caseless=True)

    def add_widget(self, child, name):
        if not isinstance(child, Menu):
            raise ValueError("child widget needs to be a Menu object")
        menu_w = self.widget.addMenu(name)
        child.widget = menu_w
        self.add_ref(child)
        self.menus[name] = child
        self.make_callback('widget-added', child)
        return child

    def add_name(self, name):
        menu_w = self.widget.addMenu(name)
        child = Menu()
        child.widget = menu_w
        self.add_ref(child)
        self.menus[name] = child
        return child

    def get_menu(self, name):
        return self.menus[name]


class TopLevelMixin(object):

    def __init__(self, title=None):

        self.widget.closeEvent = lambda event: self._quit(event)
        self.widget.destroyed = self._destroyed_cb

        if title is not None:
            self.widget.setWindowTitle(title)

        self.enable_callback('close')

    def _quit(self, event):
        # event.accept()
        # let application decide how to handle this
        event.ignore()
        self.close()

    def _closeEvent(self, *args):
        self.close()

    def close(self):
        # self.widget.deleteLater()
        # self.widget = None
        self.make_callback('close')

    def _destroyed_cb(self, event, *args):
        event.accept()

    def raise_(self):
        self.widget.raise_()
        self.widget.activateWindow()

    def lower(self):
        self.widget.lower()

    def focus(self):
        self.widget.raise_()
        self.widget.activateWindow()

    def move(self, x, y):
        self.widget.move(x, y)

    def maximize(self):
        self.widget.showMaximized()

    def unmaximize(self):
        self.widget.showNormal()

    def is_maximized(self):
        return self.widget.isMaximized()

    def fullscreen(self):
        self.widget.showFullScreen()

    def unfullscreen(self):
        self.widget.showNormal()

    def is_fullscreen(self):
        return self.widget.isFullScreen()

    def iconify(self):
        self.hide()

    def uniconify(self):
        self.widget.showNormal()

    def set_title(self, title):
        self.widget.setWindowTitle(title)


class MDIWindow(TopLevelMixin, WidgetBase):
    def __init__(self, parent, child, title='', iconpath=None):
        """NOTE: this widget is not meant to be instantiated except *inside*
        of MDIWidget implementation.
        """
        WidgetBase.__init__(self)
        self.parent = parent
        w = QtGui.QMdiSubWindow(parent.get_widget())
        # replace Qt logo from subwindow
        if iconpath is None:
            iconpath = app_icon_path
        w.setWindowIcon(QIcon(iconpath))
        #w.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.widget = w

        child_w = child.get_widget()
        # what size does the widget want to be?
        wd, ht = child_w.width(), child_w.height()
        w.setWidget(child_w)
        w.resize(wd, ht)

        # NOTE: we fire the page-switch callback by intercepting the
        # focus event on the subwindow, rather than off of the
        # subWindowActivated signal because the latter fires if
        # the widget accepts focus when the mouse enters the window,
        # whereas this approach one actually has to click in the window
        # or title bar.

        def _focus_cb(event):
            if event.gotFocus():
                self.parent._cb_redirect(self.widget)

        parent_w = parent.get_widget()
        parent_w.addSubWindow(w)

        w.focusInEvent = _focus_cb
        # Monkey-patching the widget to take control of resize and move
        # events
        w._resizeEvent = w.resizeEvent
        w.resizeEvent = lambda event: self._window_resized(event, w, child)
        w._moveEvent = w.moveEvent
        w.moveEvent = lambda event: self._window_moved(event, w, child)
        w._closeEvent = w.closeEvent
        w.closeEvent = lambda event: self._window_closed(event, w, child)
        # attach title to child
        child.extdata.tab_title = title

        TopLevelMixin.__init__(self, title=title)

        # does child have a previously saved size
        size = child.extdata.get('mdi_size', None)
        if size is not None:
            wd, ht = size
            w.resize(wd, ht)

        # does child have a previously saved position
        pos = child.extdata.get('mdi_pos', None)
        if pos is not None:
            x, y = pos
            w.move(x, y)

        child.show()
        w.show()

    def set_icon(self, iconpath):
        self.widget.setWindowIcon(QIcon(iconpath))

    def _window_resized(self, event, subwin, child):
        qsize = event.size()
        wd, ht = qsize.width(), qsize.height()
        # save size into child widget metadata
        child.extdata.mdi_size = (wd, ht)
        subwin._resizeEvent(event)

    def _window_moved(self, event, subwin, child):
        qpos = event.pos()
        x, y = qpos.x(), qpos.y()
        # save position into child widget metadata
        child.extdata.mdi_pos = (x, y)
        subwin._moveEvent(event)

    def _window_closed(self, event, subwin, widget):
        # let the application deal with this if desired in page-close
        # callback
        event.ignore()
        # self.widget.removeSubWindow(subwin)

        self.make_callback('close')


class TopLevel(TopLevelMixin, ContainerBase):

    def __init__(self, title=None, iconpath=None):
        ContainerBase.__init__(self)

        widget = QtHelp.TopLevel()

        if iconpath is None:
            iconpath = app_icon_path
        widget.setWindowIcon(QIcon(iconpath))
        self.widget = widget
        box = QtGui.QVBoxLayout()
        box.setContentsMargins(0, 0, 0, 0)
        box.setSpacing(0)
        widget.setLayout(box)

        TopLevelMixin.__init__(self, title=title)

    def set_widget(self, child):
        self.add_ref(child)
        child_w = child.get_widget()
        self.widget.layout().addWidget(child_w)

    def set_icon(self, iconpath):
        # NOTE: not guaranteed to work after the window is created because
        # this can be rendered by the window manager. Better to use the
        # constructor 'iconpath' parameter
        self.widget.setWindowIcon(QIcon(iconpath))


class Application(Callback.Callbacks):

    def __init__(self, logger=None, settings=None):
        global _app
        super(Application, self).__init__()

        self.logger = logger
        if settings is None:
            settings = Settings.SettingGroup(logger=self.logger)
        self.settings = settings
        self.settings.add_defaults(use_opengl=False)

        self.window_list = []
        self.window_dict = {}
        self.wincnt = 0
        if self.settings.get('use_opengl', False):
            # ensure we are using correct version of opengl
            # NOTE: On MacOSX w/Qt it is necessary to set the default OpenGL
            # profile BEFORE creating the QApplication object, because it
            # shares the OpenGL context
            QtHelp.set_default_opengl_context()

        # NOTE: the default value of "PassThrough" allows odd scaling
        # factors that can make text and icon rendering look terrible
        # on some platforms
        if QtHelp.have_pyqt6 or QtHelp.have_pyside6:
            QtGui.QApplication.setHighDpiScaleFactorRoundingPolicy(
                QtCore.Qt.HighDpiScaleFactorRoundingPolicy.Floor)
        #QtGui.QApplication.setAttribute(QtCore.Qt.AA_DisableHighDpiScaling, True)

        app = QtGui.QApplication([])
        # app.lastWindowClosed.connect(lambda *args: self._quit())
        self._qtapp = app
        _app = self

        # Get screen size
        screen = app.primaryScreen()
        rect = screen.availableGeometry()
        size = rect.size()
        self.screen_wd = size.width()
        self.screen_ht = size.height()

        # Get screen resolution
        xdpi = screen.physicalDotsPerInchX()
        ydpi = screen.physicalDotsPerInchY()
        self.screen_res = max(xdpi, ydpi)

        for name in ('close', 'shutdown'):
            self.enable_callback(name)

    def get_screen_size(self):
        return (self.screen_wd, self.screen_ht)

    def process_events(self):
        self._qtapp.processEvents()

    def process_end(self):
        self._qtapp.quit()

    def add_window(self, window, wid=None):
        if wid is None:
            wid = 'win%d' % (self.wincnt)
            self.wincnt += 1
        window.wid = wid
        window.url = ''
        window.app = self

        self.window_dict[wid] = window

    def get_window(self, wid):
        return self.window_dict[wid]

    def has_window(self, wid):
        return wid in self.window_dict

    def get_wids(self):
        return list(self.window_dict.keys())

    def make_window(self, title=None):
        w = TopLevel(title=title)
        self.add_window(w)
        return w

    def make_timer(self):
        return QtHelp.Timer()

    def mainloop(self):
        self._qtapp.exec()

    def close(self):
        """Called when someone is asking the application to close.
        Can register for this callback if you want an application-wide
        event to confirm closure.
        """
        self.make_callback('close')

    def quit(self):
        """Called when someone is forcibly quitting the application.
        Can register for this callback if you want an application-wide
        event to clean up before shutdown.
        """
        self.make_callback('shutdown')

        self._qtapp.quit()


class Dialog(TopLevelMixin, WidgetBase):

    def __init__(self, title='', flags=None, buttons=[],
                 parent=None, modal=False):
        WidgetBase.__init__(self)

        if parent is not None:
            parent = parent.get_widget()
        self.widget = QtGui.QDialog(parent)
        self.widget.setModal(modal)
        self.buttons = []

        vbox = QtGui.QVBoxLayout()
        vbox.setContentsMargins(0, 0, 0, 0)
        self.widget.setLayout(vbox)

        TopLevelMixin.__init__(self, title=title)

        self.content = VBox()
        vbox.addWidget(self.content.get_widget(), stretch=1)

        if len(buttons) > 0:
            hbox_w = QtGui.QWidget()
            hbox = QtGui.QHBoxLayout()
            hbox.setContentsMargins(5, 5, 5, 5)
            hbox.setSpacing(4)
            hbox_w.setLayout(hbox)

            for name, val in buttons:
                btn = Button(name)
                self.buttons.append(btn)

                def cb(val):
                    return lambda w: self._cb_redirect(val)

                btn.add_callback('activated', cb(val))
                hbox.addWidget(btn.get_widget(), stretch=1)

            vbox.addWidget(hbox_w, stretch=0)
            # self.widget.closeEvent = lambda event: self.delete()

        self.enable_callback('activated')

    def _cb_redirect(self, val):
        self.make_callback('activated', val)

    def get_content_area(self):
        return self.content


class SaveDialog(QtGui.QFileDialog):

    def __init__(self, title=None, selectedfilter=None):
        super(SaveDialog, self).__init__()

        self.title = title
        self.selectedfilter = selectedfilter
        self.widget = self

    def get_path(self):
        res = self.getSaveFileName(self, self.title, '', self.selectedfilter)

        if isinstance(res, tuple):
            res = res[0]

        if (res and self.selectedfilter is not None and
                not res.endswith(self.selectedfilter[1:])):
            res += self.selectedfilter[1:]
        if res == '':
            # user cancelled dialog
            res = None

        return res


class DragPackage(object):
    def __init__(self, src_widget):
        self.src_widget = src_widget
        self._drag = QtHelp.QDrag(self.src_widget)
        self._data = QtCore.QMimeData()
        self._drag.setMimeData(self._data)

    def set_urls(self, urls):
        _urls = [QtCore.QUrl(url) for url in urls]
        self._data.setUrls(_urls)

    def set_text(self, text):
        self._data.setText(text)

    def start_drag(self):
        self._drag.exec(QtCore.Qt.MoveAction)


# MODULE FUNCTIONS

def name_mangle(name, pfx=''):
    newname = []
    for c in name.lower():
        if not (c.isalpha() or c.isdigit() or (c == '_')):
            newname.append('_')
        else:
            newname.append(c)
    return pfx + ''.join(newname)


def hadjust(w, orientation):
    """Ostensibly, a function to reduce the vertical footprint of a widget
    that is normally used in a vertical stack (usually a Splitter), when it
    is instead used in a horizontal orientation.
    """
    if orientation != 'horizontal':
        return w
    # This currently does not seem to be needed for most plugins that are
    # coded to flow either vertically or horizontally and, in fact, reduces
    # the visual asthetic somewhat.
    ## spl = Splitter(orientation='vertical')
    ## spl.add_widget(w)
    ## spl.add_widget(Label(''))
    ## return spl
    return w


def make_widget(title, wtype):
    if wtype == 'label':
        w = Label(title)
        w.widget.setAlignment(QtCore.Qt.AlignRight)
    elif wtype == 'llabel':
        w = Label(title)
        w.widget.setAlignment(QtCore.Qt.AlignLeft)
    elif wtype in ('textentry', 'entry'):
        w = TextEntry()
        # w.widget.setMaxLength(12)
    elif wtype in ('textentryset', 'entryset'):
        w = TextEntrySet()
        # w.widget.setMaxLength(12)
    elif wtype == 'combobox':
        w = ComboBox()
    elif wtype == 'comboboxedit':
        w = ComboBox(editable=True)
    elif wtype in ('spinbox', 'spinbutton'):
        w = SpinBox(dtype=int)
    elif wtype == 'spinfloat':
        w = SpinBox(dtype=float)
    elif wtype == 'vbox':
        w = VBox()
    elif wtype == 'hbox':
        w = HBox()
    elif wtype in ('hslider', 'hscale'):
        w = Slider(orientation='horizontal')
    elif wtype in ('vslider', 'vscale'):
        w = Slider(orientation='vertical')
    elif wtype in ('checkbox', 'checkbutton'):
        w = CheckBox(title)
    elif wtype == 'radiobutton':
        w = RadioButton(title)
    elif wtype == 'togglebutton':
        w = ToggleButton(title)
    elif wtype == 'button':
        w = Button(title)
    elif wtype == 'spacer':
        w = Label('')
    elif wtype == 'textarea':
        w = TextArea(editable=True)
    elif wtype == 'toolbar':
        w = Toolbar()
    elif wtype == 'progress':
        w = ProgressBar()
    elif wtype == 'menubar':
        w = Menubar()
    elif wtype == 'dial':
        w = Dial()
    else:
        raise ValueError("Bad wtype=%s" % wtype)
    return w


def build_info(captions, orientation='vertical'):
    # numrows = len(captions)
    numcols = reduce(lambda acc, tup: max(acc, len(tup)), captions, 0)
    if (numcols % 2) != 0:
        raise ValueError("Column spec is not an even number")
    numcols = int(numcols // 2)

    widget = QtGui.QWidget()
    table = QtGui.QGridLayout()
    widget.setLayout(table)
    table.setVerticalSpacing(2)
    table.setHorizontalSpacing(4)
    table.setContentsMargins(2, 2, 2, 2)

    wb = Bunch.Bunch()
    row = 0
    for tup in captions:
        col = 0
        while col < numcols:
            idx = col * 2
            if idx < len(tup):
                title, wtype = tup[idx:idx + 2]
                if not title.endswith(':'):
                    name = name_mangle(title)
                else:
                    name = name_mangle('lbl_' + title[:-1])
                w = make_widget(title, wtype)
                table.addWidget(w.widget, row, col)
                wb[name] = w
            col += 1
        row += 1

    w = wrap(widget)
    w = hadjust(w, orientation=orientation)

    return w, wb


def wrap(native_widget):
    wrapper = WidgetBase()
    wrapper.widget = native_widget
    return wrapper

# END
